/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 0.1
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: P. Dechamboux, Y. Devaux, F. Dhumeaux.
 *
 */
package org.objectweb.speedo;

import org.objectweb.fractal.api.ComponentIdentity;
import org.objectweb.fractal.api.InterfaceReference;

import org.objectweb.fractal.adl.Launcher;
import org.objectweb.fractal.adl.Parser;

import org.objectweb.fractal.util.Fractal;
import org.objectweb.fractal.julia.control.name.NameController;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.speedo.log.Log;
import org.objectweb.speedo.runtime.mapper.JDBCMapperAttributes;
import org.objectweb.speedo.api.SpeedoProperties;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.perseus.pool.api.PoolAttributes;
import org.objectweb.perseus.cache.api.CacheAttributeController;
import org.objectweb.perseus.cache.api.CacheException;

import javax.jdo.PersistenceManagerFactory;
import javax.jdo.PersistenceManager;
import java.util.Properties;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Iterator;
import java.util.StringTokenizer;
import java.util.List;
import java.util.ArrayList;

/**
 * This class is a client helper which permits to create a new speedo
 * PersistentManagerFactory. Due to the JDO constraint the persistence manager
 * factory implementation must have a public empty contructor. But the use of
 * the fractal component (www.objectweb.org/fractal) needs to use a
 * fractal implementation to initialize components. Then this class is an
 * implementation of the PersistentManagerFactory interface which delegates all
 * calls on a persistent manager factory delegate. The constructor of this class
 * make the components creation and their initialization via Julia (fractal
 * implementation).
 *
 * @author S.Chassande-Barrioz
 */
public class Speedo implements PersistenceManagerFactory {

    public final static String CONNECTION_SPEC_FACTORY = 
        "org.objectweb.speedo.connectionSpecFactory";

    private final static String LOGGER_NAME = "org.objectweb.speedo.Speedo";

    private static Map INSTANCES = new HashMap();
    private final static String FRACTAL_PROVIDER = "fractal.provider";
    private final static String DEFAULT_FRACTAL_PROVIDER
            = "org.objectweb.fractal.julia.Julia";

    private final static String JULIA_KERNEL = "julia.kernel";
    private final static String DEFAULT_JULIA_KERNEL
            = "org.objectweb.fractal.julia.kernel.J2SEKernel";

    private final static String JULIA_CONFIG = "julia.config";
    private final static String DEFAULT_JULIA_CONFIG = "julia.cfg";

    // method called by the JDOHelper.getPersistenceManagerFactory method
    public static PersistenceManagerFactory getPersistenceManagerFactory(
            Properties props) throws Exception {
        Object instance;
        synchronized (INSTANCES) {
            instance = INSTANCES.get(props);
            if (instance == null) {
                instance = new Speedo(props);
                INSTANCES.put(props, instance);
            }
        }
        return (PersistenceManagerFactory) instance;
    }

    public static void clearInstances() {
        INSTANCES.clear();
    }

    /**
     * The persistence manager factory delegate
     */
    private PersistenceManagerFactory delegate = null;

    /**
     * It creates and initializes a real SpeedoPersistenceManagerFactory with
     * Julia (Fractal implementation).
     */
    public Speedo(Properties props) throws Exception {
        //Specify the default Fractal (Julia) if needed
        boolean isJulia = false;
        String s = System.getProperty(FRACTAL_PROVIDER);
        if (s == null || s.length() == 0) {
            isJulia = true;
            System.setProperty(FRACTAL_PROVIDER, DEFAULT_FRACTAL_PROVIDER);
        }

        if (isJulia ||
                s.equals(DEFAULT_FRACTAL_PROVIDER)) {
            s = props.getProperty(JULIA_KERNEL);
            if (s == null || s.length() == 0)
                System.setProperty(JULIA_KERNEL, DEFAULT_JULIA_KERNEL);

            s = props.getProperty(JULIA_CONFIG);
            if (s == null || s.length() == 0)
                System.setProperty(JULIA_CONFIG, DEFAULT_JULIA_CONFIG);
        }

        Parser parser = Launcher.getBootstrapParser();

        // Template of Speedo
        ComponentIdentity speedoTmpl =
                parser.loadTemplate("org.objectweb.speedo.runtime.Speedo");
        // Template of a PersistenceManagerFactory
        ComponentIdentity pmfTmpl = getSubComponent(speedoTmpl, "persistence-manager-factory");
        //Template of a PersistenceManager
        ComponentIdentity pmTmpl = getSubComponent(speedoTmpl, "persistence-manager");

        List errors = new ArrayList();

        String strval;
        //configure the pool of PersistenceManager
        PoolAttributes poolAttr = null;
        strval = props.getProperty(SpeedoProperties.PM_POOL_MAX, "");
        if (strval.length()>0) {
            try {
                if (poolAttr == null) {
                    poolAttr = getPMPoolAttributes(pmfTmpl);
                }
                poolAttr.setMaxSize(Integer.parseInt(strval));
            } catch (Exception e) {
                errors.add(new SpeedoException(
                        "Impossible to assign the max size of the pool of PersistenceManager", e));
            }
        }
        strval = props.getProperty(SpeedoProperties.PM_POOL_MIN, "");
        if (strval.length()>0) {
            try {
                if (poolAttr == null) {
                    poolAttr = getPMPoolAttributes(speedoTmpl);
                }
                poolAttr.setMinSize(Integer.parseInt(strval));
            } catch (Exception e) {
                errors.add(new SpeedoException(
                        "Impossible to assign the min size of the pool of PersistenceManager", e));
            }
        }


        //configure the cache size
        strval = props.getProperty(SpeedoProperties.CACHE_SIZE, "");
        if (strval.length()>0) {
            try {
                CacheAttributeController cacheAttr = (CacheAttributeController)
                    Fractal.getAttributeController(getSubComponent(pmTmpl,
                            "cache-manager.cache-manager"));
                if (strval.equalsIgnoreCase("nolimit")) {
                    cacheAttr.setMaxObjects(CacheAttributeController.NO_LIMIT);
                } else {
                    cacheAttr.setMaxObjects(Integer.parseInt(strval));
                }
            } catch (Exception e) {
                errors.add(new SpeedoException(
                        "Impossible to assign the size of Cache", e));
            }
        }

        //configure the pool of Query
        poolAttr = null;
        strval = props.getProperty(SpeedoProperties.QUERY_POOL_MAX, "");
        if (strval.length()>0) {
            try {
                if (poolAttr == null) {
                    poolAttr = getQueryPoolAttributes(pmTmpl);
                }
                poolAttr.setMaxSize(Integer.parseInt(strval));
            } catch (Exception e) {
                errors.add(new SpeedoException(
                        "Impossible to assign the max size of the pool of Query", e));
            }
        }
        strval = props.getProperty(SpeedoProperties.QUERY_POOL_MIN, "");
        if (strval.length()>0) {
            try {
                if (poolAttr == null) {
                    poolAttr = getQueryPoolAttributes(pmTmpl);
                }
                poolAttr.setMinSize(Integer.parseInt(strval));
            } catch (Exception e) {
                errors.add(new SpeedoException(
                        "Impossible to assign the min size of the pool of Query", e));
            }
        }

        //configure the pool of compiled query
        poolAttr = null;
        strval = props.getProperty(SpeedoProperties.COMPILED_QUERY_POOL_MAX, "");
        if (strval.length()>0) {
            try {
                if (poolAttr == null) {
                    poolAttr = getCompiledQueryPoolAttributes(pmTmpl);
                }
                poolAttr.setMaxSize(Integer.parseInt(strval));
            } catch (Exception e) {
                errors.add(new SpeedoException(
                        "Impossible to assign the max size of the pool of compiled query", e));
            }
        }
        strval = props.getProperty(SpeedoProperties.COMPILED_QUERY_POOL_MIN, "");
        if (strval.length()>0) {
            try {
                if (poolAttr == null) {
                    poolAttr = getCompiledQueryPoolAttributes(pmTmpl);
                }
                poolAttr.setMinSize(Integer.parseInt(strval));
            } catch (Exception e) {
                errors.add(new SpeedoException(
                        "Impossible to assign the min size of the pool of compiled query", e));
            }
        }

        // Instanciate the speedo template (without PM)
        Fractal.getContentController(speedoTmpl).removeFcSubComponent(pmTmpl);
        ComponentIdentity speedoComp =
                Fractal.getTemplate(speedoTmpl).instantiateFc();

        // gets the reference of the speedo persistence manager factory component,
        // which is the unique sub component of the root speedo component
        ComponentIdentity pmfComp =
                Fractal.getContentController(speedoComp).getFcSubComponents()[0];
        Fractal.getContentController(speedoComp).addFcSubComponent(pmTmpl);

        // creates the bindings between the speedo persistence manager factory
        // component and the speedo persistence manager template component
        Fractal.getBindingController(pmfComp).bindFc(
                "persistence-manager-template", (InterfaceReference) pmTmpl);
        Fractal.getBindingController(pmTmpl).bindFc(
                "persistence-manager-factory", pmfComp.getFcInterface("persistence-manager-factory"));
        Fractal.getBindingController(pmTmpl).bindFc(
                "persistence-manager-pool", pmfComp.getFcInterface("persistence-manager-pool"));

        delegate = (PersistenceManagerFactory)
                speedoComp.getFcInterface("persistence-manager-factory");

        Logger logger = ((LoggerFactory) Fractal.getContentController(pmfComp)
                .getFcSubComponents()[2].getFcInterface("logger-factory"))
                .getLogger(LOGGER_NAME);

        for(int i=0; i<errors.size(); i++) {
            Exception e = (Exception) errors.get(i);
            logger.log(BasicLevel.ERROR, e.getMessage(), e);
        }
        delegate.getProperties().putAll(props);
        boolean isJdbc = true;
        String mapperName = "rdb";
        String jdbcUrl = null;
        String jdbcUser = null;
        String jdbcPass = null;
        String jdbcDriverCN = null;
        for (Iterator it = props.entrySet().iterator(); it.hasNext();) {
            Map.Entry me = (Map.Entry) it.next();
            String key = (String) me.getKey();
            logger.log(BasicLevel.DEBUG, "(" + me.getKey() + ", " + me.getValue() + ")");
            if ("org.objectweb.speedo.mapperName".equals(key)) {
                mapperName = (String) me.getValue();

            } else if ("javax.jdo.option.DriverClassName".equals(key)) {
                jdbcDriverCN = (String) me.getValue();

            } else if ("javax.jdo.option.Optimistic".equals(key)) {
                delegate.setOptimistic(Boolean.getBoolean((String) me.getValue()));

            } else if ("javax.jdo.option.RetainValues".equals(key)) {
                delegate.setRetainValues(Boolean.getBoolean((String) me.getValue()));

            } else if ("javax.jdo.option.RestoreValues".equals(key)) {
                delegate.setRestoreValues(Boolean.getBoolean((String) me.getValue()));

            } else if ("javax.jdo.option.IgnoreCache".equals(key)) {
                delegate.setIgnoreCache(Boolean.getBoolean((String) me.getValue()));

            } else if ("javax.jdo.option.NontransactionalRead".equals(key)) {
                delegate.setNontransactionalRead(Boolean.getBoolean((String) me.getValue()));

            } else if ("javax.jdo.option.NontransactionalWrite".equals(key)) {
                delegate.setNontransactionalWrite(Boolean.getBoolean((String) me.getValue()));

            } else if ("javax.jdo.option.Multithreaded".equals(key)) {
                delegate.setMultithreaded(Boolean.getBoolean((String) me.getValue()));

            } else if ("javax.jdo.option.ConnectionUserName".equals(key)) {
                jdbcUser = (String) me.getValue();
                delegate.setConnectionUserName(jdbcUser);

            } else if ("javax.jdo.option.ConnectionPassword".equals(key)) {
                jdbcPass = (String) me.getValue();
                delegate.setConnectionPassword(jdbcPass);

            } else if ("javax.jdo.option.ConnectionURL".equals(key)) {
                jdbcUrl = (String) me.getValue();
                delegate.setConnectionURL(jdbcUrl);

            } else if ("javax.jdo.option.ConnectionFactoryName".equals(key)) {
                isJdbc = false; //The JCA mapper mus be used

            } else if ("javax.jdo.option.ConnectionFactory2Name".equals(key)) {
                //not used
            } else if ("javax.jdo.PersistenceManagerFactoryClass".equals(key)) {
                //already used to instanciate this class
            } else if (SpeedoProperties.CACHE_SIZE.equals(key)) {
                logger.log(BasicLevel.INFO, "Assign a cache size:" + me.getValue());
            } else if (SpeedoProperties.PM_POOL_MAX.equals(key)) {
                logger.log(BasicLevel.INFO, "Assign a max size of the pool of PersistenceManager:" + me.getValue());
            } else if (SpeedoProperties.PM_POOL_MIN.equals(key)) {
                logger.log(BasicLevel.INFO, "Assign a min size of the pool of PersistenceManager:" + me.getValue());
            } else if (SpeedoProperties.QUERY_POOL_MAX.equals(key)) {
                logger.log(BasicLevel.INFO, "Assign a max size of the pool of Query:" + me.getValue());
            } else if (SpeedoProperties.QUERY_POOL_MIN.equals(key)) {
                logger.log(BasicLevel.INFO, "Assign a min size of the pool of Query:" + me.getValue());
            } else if (SpeedoProperties.COMPILED_QUERY_POOL_MAX.equals(key)) {
                logger.log(BasicLevel.INFO, "Assign a max size of the pool of compiled query:" + me.getValue());
            } else if (SpeedoProperties.COMPILED_QUERY_POOL_MIN.equals(key)) {
                logger.log(BasicLevel.INFO, "Assign a min size of the pool of compiled query:" + me.getValue());
            } else
                logger.log(BasicLevel.WARN, "(" + key + ", " + me.getValue() + ") is not managed");

        }
        ComponentIdentity mapperComp =
                Fractal.getContentController(pmfComp).getFcSubComponents()[3];
        if (isJdbc) {
            JDBCMapperAttributes jdbcma = (JDBCMapperAttributes)
                    Fractal.getContentController(mapperComp).getFcSubComponents()[0].getFcInterface("attribute-controller");
            jdbcma.setMapperName(mapperName);
            jdbcma.setDriverClassName(jdbcDriverCN);
            jdbcma.setURL(jdbcUrl);
            jdbcma.setUserName(jdbcUser);
            jdbcma.setPassword(jdbcPass);
        }

        Fractal.getLifeCycleController(speedoComp).startFc();
    }

    public PoolAttributes getPMPoolAttributes(ComponentIdentity pmfTmpl) {
        return (PoolAttributes)
            Fractal.getAttributeController(getSubComponent(pmfTmpl,
                    "persistence-manager-pool"));
    }

    public PoolAttributes getQueryPoolAttributes(ComponentIdentity pmTmpl) {
        return (PoolAttributes)
            Fractal.getAttributeController(getSubComponent(pmTmpl,
                    "query-pool"));
    }

    public PoolAttributes getCompiledQueryPoolAttributes(ComponentIdentity pmTmpl) {
        return (PoolAttributes)
            Fractal.getAttributeController(getSubComponent(pmTmpl,
                    "query-manager.query-compiler-pool"));
    }

    private ComponentIdentity getSubComponent(ComponentIdentity parent,
                                              String path) {
        StringTokenizer st = new StringTokenizer(path, ".", false);
        ComponentIdentity res = parent;
        while(st.hasMoreTokens()) {
            String commponenentname = st.nextToken();
            ComponentIdentity[] children = Fractal.getContentController(res)
                    .getFcSubComponents();
            int i = 0;
            while(i<children.length &&
                    !((NameController) children[i].getFcInterface("name-controller"))
                        .getFcName().equals(commponenentname)) {
                i++;
            }
            if (i<children.length) {
                res = children[i];
            } else {
                return null;
            }
        }
        return res;
    }

    // IMPLEMENTATION OF THE PersistenceManagerFactory INTERFACE //
    //-----------------------------------------------------------//

    public PersistenceManager getPersistenceManager() {
        return delegate.getPersistenceManager();
    }

    public PersistenceManager getPersistenceManager(String s, String s1) {
        return delegate.getPersistenceManager(s, s1);
    }

    public void setConnectionUserName(String s) {
        delegate.setConnectionUserName(s);
    }

    public String getConnectionUserName() {
        return delegate.getConnectionUserName();
    }

    public void setConnectionPassword(String s) {
        delegate.setConnectionPassword(s);
    }

    public void setConnectionURL(String s) {
        delegate.setConnectionURL(s);
    }

    public String getConnectionURL() {
        return delegate.getConnectionURL();
    }

    public void setConnectionDriverName(String s) {
        delegate.setConnectionDriverName(s);
    }

    public String getConnectionDriverName() {
        return delegate.getConnectionDriverName();
    }

    public void setConnectionFactoryName(String s) {
        delegate.setConnectionFactoryName(s);
    }

    public String getConnectionFactoryName() {
        return delegate.getConnectionFactoryName();
    }

    public void setConnectionFactory(Object o) {
        delegate.setConnectionFactory(o);
    }

    public Object getConnectionFactory() {
        return delegate.getConnectionFactory();
    }

    public void setConnectionFactory2Name(String s) {
        delegate.setConnectionFactory2Name(s);
    }

    public String getConnectionFactory2Name() {
        return delegate.getConnectionFactory2Name();
    }

    public void setConnectionFactory2(Object o) {
        delegate.setConnectionFactory2(o);
    }

    public Object getConnectionFactory2() {
        return delegate.getConnectionFactory2();
    }

    public void setMultithreaded(boolean b) {
        delegate.setMultithreaded(b);
    }

    public boolean getMultithreaded() {
        return delegate.getMultithreaded();
    }

    public void setOptimistic(boolean b) {
        delegate.setOptimistic(b);
    }

    public boolean getOptimistic() {
        return delegate.getOptimistic();
    }

    public void setRetainValues(boolean b) {
        delegate.setRetainValues(b);
    }

    public boolean getRetainValues() {
        return delegate.getRetainValues();
    }

    public void setRestoreValues(boolean b) {
        delegate.setRestoreValues(b);
    }

    public boolean getRestoreValues() {
        return delegate.getRestoreValues();
    }

    public void setNontransactionalRead(boolean b) {
        delegate.setNontransactionalRead(b);
    }

    public boolean getNontransactionalRead() {
        return delegate.getNontransactionalRead();
    }

    public void setNontransactionalWrite(boolean b) {
        delegate.setNontransactionalWrite(b);
    }

    public boolean getNontransactionalWrite() {
        return delegate.getNontransactionalWrite();
    }

    public void setIgnoreCache(boolean b) {
        delegate.setIgnoreCache(b);
    }

    public boolean getIgnoreCache() {
        return delegate.getIgnoreCache();
    }

    public int getMaxPool() {
        return delegate.getMaxPool();
    }

    public void setMaxPool(int i) {
        delegate.setMaxPool(i);
    }

    public int getMinPool() {
        return delegate.getMinPool();
    }

    public void setMinPool(int i) {
        delegate.setMinPool(i);
    }

    public int getMsWait() {
        return delegate.getMsWait();
    }

    public void setMsWait(int i) {
        delegate.setMsWait(i);
    }

    public Properties getProperties() {
        return delegate.getProperties();
    }

    public Collection supportedOptions() {
        return delegate.supportedOptions();
    }
}
