/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 0.1
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: P. Dechamboux, Y. Devaux, F. Dhumeaux.
 *
 */

package org.objectweb.speedo.ant;

import org.apache.tools.ant.BuildException;
import org.apache.tools.ant.DirectoryScanner;
import org.apache.tools.ant.Project;
import org.apache.tools.ant.Task;
import org.apache.tools.ant.taskdefs.MatchingTask;
import org.apache.tools.ant.taskdefs.Javac;
import org.apache.tools.ant.types.DTDLocation;
import org.apache.tools.ant.types.Path;
import org.apache.tools.ant.types.Reference;
import org.objectweb.jorm.api.PException;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.speedo.api.SpeedoProxy;
import org.objectweb.speedo.generation.SpeedoCompiler;
import org.objectweb.speedo.generation.api.SpeedoCompilerParameter;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * Ant task that can be used to execute the generation.
 * <p>Parameters of the task:</p><ul>
 * <li><code>localpropertiesfile</code>, indicates whether there is a jdo.properties</li>
 * <li><code>confFile</code>, location of the corresponding properties files</li>
 * <li><code>logPropFile</code>, location of the log system properties file</li>
 * <li><code>projectname</code> is the project name (optional) (Jorm parameter)</li>
 * <li><code>mappername</code> is the mapper name for which the code will be generated</li>
 * <li><code>keepsrc</code>, indicates whether generated Java files must be kept (optionnal, the default value is true</li>
 * <li><code>input</code>, location of the base directory of <code>.class</code> files</li>
 * <li><code>output</code>, location of the base destination directory</li>
 * <li><code>classpath</code> nested classpath element containing the speedo and jorm jars</li>
 * <li><code>jdopath</code> nested element containing a parameter <code>dir</code> and <code>include elements</code> , location of <code>.jdo</code> JDO XML Descriptors</li>
 * <li><code>jormpath</code> (similar to <code>jdopath</code>), location of <code>.pd</code> JORM XML Descriptors</li>
 * <li><code>persistenceaware</code> (similar to <code>jdopath</code>), location of Persistence Aware Java Class</li>
 * <li><code>keepclasses</code>, indicates whether original java classes must be stored in another directory (see below)</li>
 * <li><code>storedir</code>, location for storing original classes</li>
 * <li><code>failsonerror</code>, indicates whether the process should fail if an error occurs</li></ul>
 * @author	Y. Devaux
 */
public class AntSpeedo extends Task {

    private Path classpath = null;
    private Path jormclasspath = null;
    private Description jdo = null;
    private Description jorm = null;
    private Description awareFiles = null;

    private SpeedoCompilerParameter scp = null;
    private SpeedoCompiler sc = null;
    private File src = null;
    private File output = null;

    public void init() throws BuildException {
        super.init();
        sc = new SpeedoCompiler();
        scp = sc.getSpeedoCompilerParameter();
    }

    public void setConfFile(File jormconfFile) {
        scp.confFile = jormconfFile.getAbsolutePath();
        try {
            scp.loadConfigFile();
        } catch (IOException e) {
            throw new BuildException(
                    "Impossible to load the configuration file", e);
        }
    }

    public void setLocalpropertiesfile(boolean jormlocalpropertiesfile) {
        scp.localpropertiesfile = jormlocalpropertiesfile;
    }

    public void setLogPropFile(File logPropFile) {
        scp.logPropFile = logPropFile.getAbsolutePath();
    }

    public void setProjectname(String projectName) {
        scp.projectName = projectName;
    }

    public void setMappername(String mapperName) {
        scp.mapperName = mapperName;
    }

    public void setSrc(File src) {
        this.src = src;
    }

    public void setKeepsrc(boolean keepsrc) {
        scp.keepsrc = keepsrc;
    }

    public void setFailsonerror(boolean fail) {
        scp.failsonerror = fail;
    }

    /**
     * Create a DTD location record. This stores the location of a DTD. The
     * DTD is identified by its public Id. The location may either be a file
     * location or a resource location.
     *
     * @return the DTD location object to be configured by Ant
     */
    public DTDLocation createDTD() {
        DTDLocation dtdLocation = new DTDLocation();
        scp.dtdLocations.add(dtdLocation);
        return dtdLocation;
    }

    public void setOutput(File output) {
        this.output = output;
        scp.output = output.getAbsolutePath();
    }

    public void setInput(File intput) {
        scp.input = intput.getAbsolutePath();
    }

    public void setClasspath(Path cp) {
        if (classpath == null)
            classpath = cp;
        else
            classpath.append(cp);
    }

    public Path getClasspath() {
        return classpath;
    }

    public Path createClasspath() {
        if (classpath == null)
            classpath = new Path(project);
        return classpath.createPath();
    }

    public void setClasspathRef(Reference r) {
        createClasspath().setRefid(r);
    }

    public Description createJdopath() {
        Description d = new Description();
        this.jdo = d;
        d.setProject(getProject());
        return d;
    }

    public Description createJormpath() {
        Description d = new Description();
        this.jorm = d;
        d.setProject(getProject());
        return d;
    }

    public void setJormClasspath(Path cp) {
        if (jormclasspath == null)
            jormclasspath = cp;
        else
            jormclasspath.append(cp);
    }

    public Path getJormClasspath() {
        return jormclasspath;
    }

    public Path createJormClasspath() {
        if (jormclasspath == null)
            jormclasspath = new Path(project);
        return jormclasspath.createPath();
    }

    public void setJormClasspathRef(Reference r) {
        createJormClasspath().setRefid(r);
    }

    public Description createPersistenceaware() {
        Description d = new Description();
        this.awareFiles = d;
        d.setProject(getProject());
        return d;
    }

    public void setKeepclasses(boolean val) {
        scp.keepclasses = val;
    }


    /**
     * Main method of the task executed by ant.
     * Parses xml parameters, loads AntSpeedoExec with its own loader
     */
    public void execute() throws BuildException {
        if (jdo == null) {
            jdo = createJdopath();
            jdo.setDir(src);
            jdo.setIncludes("**/*.jdo");
        }
        scp.jdoDir = jdo.dir.getAbsolutePath();
        scp.jdo = new MyArrayList(
                jdo.getDirectoryScanner(jdo.dir).getIncludedFiles());


        if (jorm == null) {
            jorm = createJormpath();
            jorm.setDir(src);
            jorm.setIncludes("**/*.pd");
        }
        scp.jormDir = jorm.dir.getAbsolutePath();
        scp.jorm = new MyArrayList(
                jorm.getDirectoryScanner(jorm.dir).getIncludedFiles());

        if (awareFiles == null) {
            awareFiles = createJdopath();
            awareFiles.setDir(output);
            awareFiles.setIncludes("**/*.class");
        }
        scp.awareFilesDir = awareFiles.dir.getAbsolutePath();
        scp.awareFiles = new MyArrayList(
                awareFiles.getDirectoryScanner(awareFiles.dir).getIncludedFiles());

        if (jormclasspath == null) {
            jormclasspath = classpath;
        }
        scp.jormclasspath = new MyArrayList(jormclasspath.list());
        scp.classpath = classpath;

        log("scp.jormclasspath=" + scp.jormclasspath, Project.MSG_DEBUG);

        if (scp.dtdLocations.size() == 0) {
            DTDLocation dl = new DTDLocation();
            dl.setPublicId("-//ObjectWeb Consortium//DTD JORM DEFINITIONS 2.0//EN");
            dl.setLocation("jorm2.dtd");
            scp.dtdLocations.add(dl);
            dl = new DTDLocation();
            dl.setPublicId("-//Sun Microsystems, Inc.//DTD Java Data Objects Metadata 1.0//EN");
            dl.setLocation("jdo.dtd");
            scp.dtdLocations.add(dl);
        }

        if (scp.input == null || scp.input.length()==0) {
            scp.input = scp.output;
        }

        File dir = new File(scp.output);
        List sourceToCompile = null;
        if (dir.exists()) {
            sourceToCompile = new ArrayList();
            log("Remove the specified persistent classes from the directory "
                    + scp.output, Project.MSG_INFO);
            for (Iterator it = scp.jorm.iterator(); it.hasNext();) {
                String fn = (String) it.next();
                if (!fn.endsWith(".pd")) {
                    log("You specify a file which is not a Jorm file(.pd): " + fn, Project.MSG_WARN);
                    it.remove();
                    continue;
                }
                fn = fn.substring(0, fn.length() - 3);
                String cn = fn.replace(File.separatorChar, '.');
                String jfn = fn + ".java";
                String cfn = fn + ".class";
                File f = new File(dir, jfn);
                f.delete();
                f = new File(dir, cfn);
                if (f.exists()) {
                    try {
                        if (new MyArrayList(
                                Class.forName(cn).getInterfaces())
                                .contains(SpeedoProxy.class)) {
                            log("Delete the class " + cn, Project.MSG_INFO);
                            f.delete();
                            sourceToCompile.add(jfn);
                        }
                    } catch (ClassNotFoundException e) {
                        sourceToCompile.add(jfn);
                    } catch (Throwable e) {
                        log("Error in class loading: " + e.getMessage(), Project.MSG_ERR);
                        log("Delete the class " + cn, Project.MSG_INFO);
                        f.delete();
                        sourceToCompile.add(jfn);
                    }
                } else {
                    sourceToCompile.add(jfn);
                    log("Class " + cn + " does not exist", Project.MSG_DEBUG);
                }
            }

            log("Compile", Project.MSG_DEBUG);
            Javac javac = new Javac();
            javac.setProject(project);
            javac.setTaskName(taskName);
            Path p = new Path(project);
            p.setPath(scp.input);
            javac.createSrc().createPathElement().setLocation(src);

            javac.setDestdir(new File(scp.output));
            javac.setClasspath(classpath);
            javac.setDebug(true);
            //javac.setDebugLevel("lines,vars,source");
            javac.setFailonerror(true);
            javac.createInclude().setName("**/*.java");
            for (Iterator it = sourceToCompile.iterator(); it.hasNext();) {
                String fn = (String) it.next();
                log("Compile " + fn, Project.MSG_DEBUG);
                
                javac.createInclude().setName(fn);
            }
            javac.execute();

            log("Remove the jmi files from the directory " + scp.output,
                    Project.MSG_INFO);
            for (Iterator it = scp.jdo.iterator(); it.hasNext();) {
                String fn = (String) it.next();
                if (!fn.endsWith(".jdo")) {
                    log("You specify a file which is not a jdo file(.jdo): "
                            + fn, Project.MSG_WARN);
                    it.remove();
                    continue;
                }
                fn = fn.substring(0, fn.length() - 3) + "jmi";
                File f = new File(dir, fn);
                if (f.exists()) {
                    log("Delete the jmi file " + fn, Project.MSG_DEBUG);
                    f.delete();
                } else
                    log("The jmi file " + fn + " does not exist", Project.MSG_DEBUG);
            }

        }

        try {
            sc.init();
            sc.process();
        } catch (SpeedoException e) {
            e.printStackTrace();
            throw new BuildException(getNestedException(e));
        }
    }

    private Exception getNestedException(Exception e) {
        if (e instanceof SpeedoException
                && ((SpeedoException) e).getNestedException() != null) {
            return getNestedException(((SpeedoException) e).getNestedException());
        } else if (e instanceof PException
                && ((PException) e).getNestedException() != null) {
            return getNestedException(((PException) e).getNestedException());
        } else
            return e;
    }

    /**
     * Task used to parse nested <code>jdopath</code> or <code>jormpath</code> elements.
     */
    public class Description extends MatchingTask {

        public File dir = null;

        public void setDir(File dir) {
            this.dir = dir;
        }

        public DirectoryScanner getDirectoryScanner(File p) {
            return super.getDirectoryScanner(p);
        }
    }

}

class MyArrayList extends ArrayList {
    public MyArrayList(Object[] os) {
        super(os.length);
        for (int i = 0; i < os.length; i++)
            add(os[i]);
    }
}
