/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 0.1
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: P. Dechamboux, Y. Devaux, F. Dhumeaux.
 *
 */

package org.objectweb.speedo.api;

/**
 * Representation of the life cycle of a JDO Instance.
 * @author	Y. Devaux
 */
public abstract class LifeCycle {
    public static final byte NOT_APPLICABLE = -2;
    public static final byte ERROR = -1;
    public static final byte TRANSIENT = 0;
    public static final byte TRANSIENT_CLEAN = 1;
    public static final byte TRANSIENT_DIRTY = 2;
    public static final byte PERSISTENT_NEW = 3;
    public static final byte PERSISTENT_NONTRANSACTIONAL = 4;
    public static final byte PERSISTENT_CLEAN = 5;
    public static final byte PERSISTENT_DIRTY = 6;
    public static final byte HOLLOW = 7;
    public static final byte PERSISTENT_DELETED = 8;
    public static final byte PERSISTENT_NEW_DELETED = 9;

    public static final byte ACTION_MAKEPERSISTENT = 0;
    public static final byte ACTION_DELETEPERSISTENT = 1;
    public static final byte ACTION_MAKETRANSACTIONAL = 2;
    public static final byte ACTION_MAKENONTRANSACTIONAL = 3;
    public static final byte ACTION_MAKETRANSIENT = 4;
    public static final byte ACTION_COMMIT = 5;
    public static final byte ACTION_COMMIT_RETAINVALUES = 6;
    public static final byte ACTION_ROLLBACK = 7;
    public static final byte ACTION_ROLLBACK_RETAINVALUES = 8;
    public static final byte ACTION_REFRESH_ACTIVEDATASTORETRANSACTION = 9;
    public static final byte ACTION_REFRESH_ACTIVEOPTIMISTICTRANSACTION = 10;
    public static final byte ACTION_EVICT = 11;
    public static final byte ACTION_READFIELD_OUTSIDETRANSACTION = 12;
    public static final byte ACTION_READFIELD_ACTIVEOPTIMISTICTRANSACTION = 13;
    public static final byte ACTION_READFIELD_ACTIVEDATASTORETRANSACTION = 14;
    public static final byte ACTION_WRITEFIELD_ACTIVEDATASTORETRANSACTION = 15;
    public static final byte ACTION_WRITEFIELD_OUTSIDETRANSACTION = 16;
    // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    public static final byte ACTION_MAKEDIRTY = 15;
    // !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

    /**
     * Gets the state transition in the JDO instances life cycle
     * @param status current instance status
     * @param action method that is currently invoked
     * @return the next state according to JDO rules
     */
    public static byte nextStatePersistenceCapable(byte status, byte action) {
        switch (status) {
        case (TRANSIENT):
            switch (action) {
            case (ACTION_MAKEPERSISTENT):
                return PERSISTENT_NEW;
            case (ACTION_MAKETRANSACTIONAL):
                return TRANSIENT_CLEAN;
            case (ACTION_MAKETRANSIENT):
            case (ACTION_COMMIT):
            case (ACTION_COMMIT_RETAINVALUES):
            case (ACTION_ROLLBACK):
            case (ACTION_ROLLBACK_RETAINVALUES):
            case (ACTION_READFIELD_OUTSIDETRANSACTION):
            case (ACTION_READFIELD_ACTIVEOPTIMISTICTRANSACTION):
            case (ACTION_READFIELD_ACTIVEDATASTORETRANSACTION):
            case (ACTION_WRITEFIELD_ACTIVEDATASTORETRANSACTION):
            case (ACTION_WRITEFIELD_OUTSIDETRANSACTION):
                return TRANSIENT;
            case (ACTION_MAKENONTRANSACTIONAL):
            case (ACTION_REFRESH_ACTIVEDATASTORETRANSACTION):
            case (ACTION_REFRESH_ACTIVEOPTIMISTICTRANSACTION):
            case (ACTION_EVICT):
                return NOT_APPLICABLE;
            case (ACTION_DELETEPERSISTENT):
            default :
                return ERROR;
            }
        case (TRANSIENT_CLEAN):
            switch (action) {
            case (ACTION_MAKEPERSISTENT):
                return PERSISTENT_NEW;
            case (ACTION_MAKENONTRANSACTIONAL):
                return TRANSIENT;
            case (ACTION_WRITEFIELD_OUTSIDETRANSACTION):
                return TRANSIENT_DIRTY;
            case (ACTION_MAKETRANSACTIONAL):
            case (ACTION_MAKETRANSIENT):
            case (ACTION_COMMIT):
            case (ACTION_COMMIT_RETAINVALUES):
            case (ACTION_ROLLBACK):
            case (ACTION_ROLLBACK_RETAINVALUES):
            case (ACTION_EVICT):
            case (ACTION_READFIELD_OUTSIDETRANSACTION):
            case (ACTION_READFIELD_ACTIVEOPTIMISTICTRANSACTION):
            case (ACTION_READFIELD_ACTIVEDATASTORETRANSACTION):
            case (ACTION_WRITEFIELD_ACTIVEDATASTORETRANSACTION):
                return TRANSIENT_CLEAN;
            case (ACTION_REFRESH_ACTIVEDATASTORETRANSACTION):
            case (ACTION_REFRESH_ACTIVEOPTIMISTICTRANSACTION):
                return NOT_APPLICABLE;
            case (ACTION_DELETEPERSISTENT):
            default :
                return ERROR;
            }
        case (TRANSIENT_DIRTY):
            switch (action) {
            case (ACTION_MAKEPERSISTENT):
                return PERSISTENT_NEW;
            case (ACTION_COMMIT):
            case (ACTION_COMMIT_RETAINVALUES):
            case (ACTION_ROLLBACK):
            case (ACTION_ROLLBACK_RETAINVALUES):
                return TRANSIENT_CLEAN;
            case (ACTION_MAKETRANSACTIONAL):
            case (ACTION_MAKETRANSIENT):
            case (ACTION_EVICT):
            case (ACTION_READFIELD_OUTSIDETRANSACTION):
            case (ACTION_READFIELD_ACTIVEOPTIMISTICTRANSACTION):
            case (ACTION_READFIELD_ACTIVEDATASTORETRANSACTION):
            case (ACTION_WRITEFIELD_ACTIVEDATASTORETRANSACTION):
                return TRANSIENT_DIRTY;
            case (ACTION_REFRESH_ACTIVEDATASTORETRANSACTION):
            case (ACTION_REFRESH_ACTIVEOPTIMISTICTRANSACTION):
            case (ACTION_WRITEFIELD_OUTSIDETRANSACTION):
                return NOT_APPLICABLE;
            case (ACTION_DELETEPERSISTENT):
            case (ACTION_MAKENONTRANSACTIONAL):
            default :
                return ERROR;
            }
        case (PERSISTENT_NEW):
            switch (action) {
            case (ACTION_DELETEPERSISTENT):
                return PERSISTENT_NEW_DELETED;
            case (ACTION_COMMIT):
                return HOLLOW;
            case (ACTION_COMMIT_RETAINVALUES):
                return PERSISTENT_NONTRANSACTIONAL;
            case (ACTION_ROLLBACK):
            case (ACTION_ROLLBACK_RETAINVALUES):
                return TRANSIENT;
            case (ACTION_MAKEPERSISTENT):
            case (ACTION_MAKETRANSACTIONAL):
            case (ACTION_REFRESH_ACTIVEDATASTORETRANSACTION):
            case (ACTION_REFRESH_ACTIVEOPTIMISTICTRANSACTION):
            case (ACTION_EVICT):
            case (ACTION_READFIELD_OUTSIDETRANSACTION):
            case (ACTION_READFIELD_ACTIVEOPTIMISTICTRANSACTION):
            case (ACTION_READFIELD_ACTIVEDATASTORETRANSACTION):
            case (ACTION_WRITEFIELD_ACTIVEDATASTORETRANSACTION):
                return PERSISTENT_NEW;
            case (ACTION_MAKENONTRANSACTIONAL):
            case (ACTION_WRITEFIELD_OUTSIDETRANSACTION):
                return NOT_APPLICABLE;
            case (ACTION_MAKETRANSIENT):
            default :
                return ERROR;
            }
        case (PERSISTENT_NONTRANSACTIONAL):
            switch (action) {
            case (ACTION_DELETEPERSISTENT):
                return PERSISTENT_DELETED;
            case (ACTION_MAKETRANSACTIONAL):
            case (ACTION_READFIELD_ACTIVEDATASTORETRANSACTION):
                return PERSISTENT_CLEAN;
            case (ACTION_MAKETRANSIENT):
                return TRANSIENT;
            case (ACTION_EVICT):
                return HOLLOW;
            case (ACTION_WRITEFIELD_ACTIVEDATASTORETRANSACTION):
                return PERSISTENT_DIRTY;
            case (ACTION_MAKEPERSISTENT):
            case (ACTION_MAKENONTRANSACTIONAL):
            case (ACTION_COMMIT):
            case (ACTION_COMMIT_RETAINVALUES):
            case (ACTION_ROLLBACK):
            case (ACTION_ROLLBACK_RETAINVALUES):
            case (ACTION_REFRESH_ACTIVEDATASTORETRANSACTION):
            case (ACTION_REFRESH_ACTIVEOPTIMISTICTRANSACTION):
            case (ACTION_READFIELD_OUTSIDETRANSACTION):
            case (ACTION_READFIELD_ACTIVEOPTIMISTICTRANSACTION):
            case (ACTION_WRITEFIELD_OUTSIDETRANSACTION):
                return PERSISTENT_NONTRANSACTIONAL;
            default :
                return ERROR;
            }
        case (PERSISTENT_CLEAN):
            switch (action) {
            case (ACTION_DELETEPERSISTENT):
                return PERSISTENT_DELETED;
            case (ACTION_MAKENONTRANSACTIONAL):
                return PERSISTENT_NONTRANSACTIONAL;
            case (ACTION_MAKETRANSIENT):
                return TRANSIENT;
            case (ACTION_COMMIT):
            case (ACTION_ROLLBACK):
            case (ACTION_EVICT):
                return HOLLOW;
            case (ACTION_COMMIT_RETAINVALUES):
            case (ACTION_ROLLBACK_RETAINVALUES):
                return PERSISTENT_NONTRANSACTIONAL;
            case (ACTION_WRITEFIELD_ACTIVEDATASTORETRANSACTION):
                return PERSISTENT_DIRTY;
            case (ACTION_MAKEPERSISTENT):
            case (ACTION_MAKETRANSACTIONAL):
            case (ACTION_REFRESH_ACTIVEDATASTORETRANSACTION):
            case (ACTION_REFRESH_ACTIVEOPTIMISTICTRANSACTION):
            case (ACTION_READFIELD_OUTSIDETRANSACTION):
            case (ACTION_READFIELD_ACTIVEOPTIMISTICTRANSACTION):
            case (ACTION_READFIELD_ACTIVEDATASTORETRANSACTION):
                return PERSISTENT_CLEAN;
            case (ACTION_WRITEFIELD_OUTSIDETRANSACTION):
                return NOT_APPLICABLE;
            default :
                return ERROR;
            }
        case (PERSISTENT_DIRTY):
            switch (action) {
            case (ACTION_DELETEPERSISTENT):
                return PERSISTENT_DELETED;
            case (ACTION_COMMIT):
            case (ACTION_ROLLBACK):
                return HOLLOW;
            case (ACTION_COMMIT_RETAINVALUES):
            case (ACTION_ROLLBACK_RETAINVALUES):
            case (ACTION_REFRESH_ACTIVEOPTIMISTICTRANSACTION):
                return PERSISTENT_NONTRANSACTIONAL;
            case (ACTION_REFRESH_ACTIVEDATASTORETRANSACTION):
                return PERSISTENT_CLEAN;
            case (ACTION_MAKEPERSISTENT):
            case (ACTION_MAKETRANSACTIONAL):
            case (ACTION_EVICT):
            case (ACTION_READFIELD_OUTSIDETRANSACTION):
            case (ACTION_READFIELD_ACTIVEOPTIMISTICTRANSACTION):
            case (ACTION_READFIELD_ACTIVEDATASTORETRANSACTION):
            case (ACTION_WRITEFIELD_ACTIVEDATASTORETRANSACTION):
                return PERSISTENT_DIRTY;
            case (ACTION_MAKENONTRANSACTIONAL):
            case (ACTION_WRITEFIELD_OUTSIDETRANSACTION):
                return NOT_APPLICABLE;
            case (ACTION_MAKETRANSIENT):
            default :
                return ERROR;
            }
        case (HOLLOW):
            switch (action) {
            case (ACTION_DELETEPERSISTENT):
                return PERSISTENT_DELETED;
            case (ACTION_MAKETRANSIENT):
                return TRANSIENT;
            case (ACTION_READFIELD_OUTSIDETRANSACTION):
            case (ACTION_READFIELD_ACTIVEOPTIMISTICTRANSACTION):
            case (ACTION_WRITEFIELD_OUTSIDETRANSACTION):
                return PERSISTENT_NONTRANSACTIONAL;
            case (ACTION_READFIELD_ACTIVEDATASTORETRANSACTION):
                return PERSISTENT_CLEAN;
            case (ACTION_WRITEFIELD_ACTIVEDATASTORETRANSACTION):
                return PERSISTENT_DIRTY;
            case (ACTION_MAKEPERSISTENT):
            case (ACTION_MAKETRANSACTIONAL):
            case (ACTION_MAKENONTRANSACTIONAL):
            case (ACTION_COMMIT):
            case (ACTION_COMMIT_RETAINVALUES):
            case (ACTION_ROLLBACK):
            case (ACTION_ROLLBACK_RETAINVALUES):
            case (ACTION_REFRESH_ACTIVEDATASTORETRANSACTION):
            case (ACTION_REFRESH_ACTIVEOPTIMISTICTRANSACTION):
                ;
            case (ACTION_EVICT):
                return HOLLOW;
            default :
                return ERROR;
            }
        case (PERSISTENT_DELETED):
            switch (action) {
            case (ACTION_COMMIT):
            case (ACTION_COMMIT_RETAINVALUES):
                return TRANSIENT;
            case (ACTION_ROLLBACK):
                return HOLLOW;
            case (ACTION_ROLLBACK_RETAINVALUES):
                return PERSISTENT_NONTRANSACTIONAL;
            case (ACTION_MAKEPERSISTENT):
            case (ACTION_DELETEPERSISTENT):
            case (ACTION_MAKETRANSACTIONAL):
            case (ACTION_REFRESH_ACTIVEDATASTORETRANSACTION):
            case (ACTION_REFRESH_ACTIVEOPTIMISTICTRANSACTION):
            case (ACTION_EVICT):
                return PERSISTENT_DELETED;
            case (ACTION_MAKENONTRANSACTIONAL):
            case (ACTION_MAKETRANSIENT):
            case (ACTION_READFIELD_OUTSIDETRANSACTION):
            case (ACTION_READFIELD_ACTIVEOPTIMISTICTRANSACTION):
            case (ACTION_READFIELD_ACTIVEDATASTORETRANSACTION):
            case (ACTION_WRITEFIELD_ACTIVEDATASTORETRANSACTION):
            case (ACTION_WRITEFIELD_OUTSIDETRANSACTION):
            default :
                return ERROR;
            }
        case (PERSISTENT_NEW_DELETED):
            switch (action) {
            case (ACTION_COMMIT):
            case (ACTION_COMMIT_RETAINVALUES):
            case (ACTION_ROLLBACK):
            case (ACTION_ROLLBACK_RETAINVALUES):
                return TRANSIENT;
            case (ACTION_MAKEPERSISTENT):
            case (ACTION_DELETEPERSISTENT):
            case (ACTION_MAKETRANSACTIONAL):
            case (ACTION_REFRESH_ACTIVEDATASTORETRANSACTION):
            case (ACTION_REFRESH_ACTIVEOPTIMISTICTRANSACTION):
            case (ACTION_EVICT):
                return PERSISTENT_NEW_DELETED;
            case (ACTION_MAKENONTRANSACTIONAL):
            case (ACTION_MAKETRANSIENT):
            case (ACTION_READFIELD_OUTSIDETRANSACTION):
            case (ACTION_READFIELD_ACTIVEOPTIMISTICTRANSACTION):
            case (ACTION_READFIELD_ACTIVEDATASTORETRANSACTION):
            case (ACTION_WRITEFIELD_ACTIVEDATASTORETRANSACTION):
            case (ACTION_WRITEFIELD_OUTSIDETRANSACTION):
            default :
                return ERROR;
            }
        default :
            return ERROR;
        }
    }

    /**
     * Make the instance dirty.
     * @param status current status
     * @return new status
     */
    public static byte makeDirty(byte status) {
        switch (status) {
        case (TRANSIENT):
            return TRANSIENT;
        case (PERSISTENT_NEW):
            return PERSISTENT_NEW;
        case (PERSISTENT_CLEAN):
        case (PERSISTENT_DIRTY):
        case (HOLLOW):
        case (PERSISTENT_NONTRANSACTIONAL):
            return PERSISTENT_DIRTY;
        case (TRANSIENT_CLEAN):
        case (TRANSIENT_DIRTY):
            return TRANSIENT_DIRTY;
        case (PERSISTENT_NEW_DELETED):
            return PERSISTENT_NEW_DELETED;
        case (PERSISTENT_DELETED):
            return PERSISTENT_DELETED;
        }
        return -1;
    }

    /**
     * Indicates if the instance is persistent
     *
     * @param status current instance status
     * @return true if the instance is persistent
     */
    public static boolean isPersistent(byte status) {
        return status == PERSISTENT_NEW
                || status == PERSISTENT_CLEAN
                || status == PERSISTENT_DIRTY
                || status == HOLLOW
                || status == PERSISTENT_NONTRANSACTIONAL
                || status == PERSISTENT_NEW_DELETED
                || status == PERSISTENT_DELETED;
    }

    /**
     * Indicates if the instance is transactional
     *
     * @param status current instance status
     * @return true if the instance is transactional
     */
    public static boolean isTransactional(byte status) {
        return status == PERSISTENT_NEW
                || status == TRANSIENT_CLEAN
                || status == TRANSIENT_DIRTY
                || status == PERSISTENT_CLEAN
                || status == PERSISTENT_DIRTY
                || status == PERSISTENT_NEW_DELETED
                || status == PERSISTENT_DELETED;
    }

    /**
     * Indicates if the instance is new
     *
     * @param status current instance status
     * @return true if the instance is bew
     */
    public static boolean isNew(byte status) {
        return status == PERSISTENT_NEW || status == PERSISTENT_NEW_DELETED;
    }

    /**
     * Indicates if the instance is dirty
     *
     * @param status current instance status
     * @return true if the instance is dirty
     */
    public static boolean isDirty(byte status) {
        return status == PERSISTENT_NEW
                || status == TRANSIENT_DIRTY
                || status == PERSISTENT_DIRTY
                || status == PERSISTENT_NEW_DELETED
                || status == PERSISTENT_DELETED;
    }

    /**
     * Indicates if the instance is deleted
     *
     * @param status current instance status
     * @return true if the instance is deleted
     */
    public static boolean isDeleted(byte status) {
        return status == PERSISTENT_NEW_DELETED || status == PERSISTENT_DELETED;
    }

    /**
     * Gets the initial state
     *
     * @return <code>TRANSIENT</code>
     */
    public static byte initState() {
        return TRANSIENT;
    }

    /**
     * Indicates if the instance is transient
     *
     * @param status current instance status
     * @return true if the instance is transient
     */
    public static boolean isTransient(byte status) {
        return status == TRANSIENT
                || status == TRANSIENT_CLEAN
                || status == TRANSIENT_DIRTY;
    }
}
