/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 0.1
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: P. Dechamboux, Y. Devaux, F. Dhumeaux.
 *
 */

package org.objectweb.speedo.api;

import org.objectweb.jorm.api.PBinding;
import org.objectweb.perseus.persistence.api.TransactionalPersistenceManager;

import javax.jdo.spi.PersistenceCapable;
import javax.jdo.PersistenceManagerFactory;
import java.util.Collection;

/**
 * This interface what is a Proxy in Speedo. A proxy a PersistentCapable
 * instance because it is the user object. It is the jorm PBinding because it is
 * unique even in optimistic case. Finally it is also a SpeedoAccessor because
 * in pessimistic mode or when the object is not active the state values are
 * stored inside the proxy. In addition there are several fields: <UL>
 * <LI>The proxy state is defined and managed by a status field.</LI>
 * <LI>The field 'isActive' indicates if the proxy is active (linked to a ProxyManager)</LI>
 * <LI>The field 'isModified' indicates if the proxy has been modified since its
 * last writing.</LI>
 * <LI>The fields 'PersistentManagerFactory' represents the data source where
 * the current object is persistent.</LI>
 * <LI>The fields 'ProxyManagerSwitch' which maintains the association thread,
 * proxy manager instances.</LI>
 * </UL>
 *
 * @author	S.Chassande-Barrioz
 */
public interface SpeedoProxy
        extends PersistenceCapable, PBinding, SpeedoAccessor {

    /**
     * Stores fields owned by this object into an accessor
     *
     * @param sa the accessor requiring persistent fields
     */
    void jdoLoadFieldsInAccessor(SpeedoAccessor sa);

    /**
     * Stores fields owned by the accessor into the proxy
     *
     * @param sa the accessor containing persistent fields
     */
    void jdoLoadFieldsFromAccessor(SpeedoAccessor sa);


    /**
     * Tests whether the Proxy infrastructure has already been set.
     * A proxy has a correct PClassMapping linked to a PBinder,
     * a PName obtained thanks to this PClassMapping.
     *
     * @return true if this object is a Proxy.
     */
    boolean jdoIsActive();

    /**
     * Sets the new value of the proxy flags
     *
     * @see #jdoIsActive()
     * @param newvalue the new value of the proxy flag
     */
    void jdoIsActive(boolean newvalue);


    /**
     * It retrieves the SpeedoAccessor instance used in the current context.
     * If the proxy is not active then 'this' is returned.
     * If there is an active and otpimistic transaction then a $classNameFields
     * is returned. Otherwise 'this' is returned.
     */
    SpeedoAccessor getSpeedoAccessor();

    /**
     * Removes the given speedo accessor from the list of accessors of this
     * proxy.
     * @param sa a speedo accessor of this proxy.
     */
    void removeSpeedoAccessor(SpeedoAccessor sa);

    /**
     * It retrieves the SpeedoAccessor instance used in the current context.
     * If the proxy is not active then 'this' is returned.
     * If there is an active and otpimistic transaction then a $classNameFields
     * is returned. Otherwise 'this' is returned.
     */
    Collection getSpeedoAccessors();

    /**
     * Indicates the identity-type of this object
     *
     * @return true if the object lets the JDO Implementation manages its
     * identity, false otherwise (application identity)
     */
    boolean jdoIsContainerId();


    /**
     * @return the ProxyManager switch links to the proxy
     */
    ProxyManagerSwitch jdoGetPersistenceManagerSwitch();

    /**
     * It assignes the ProxyManager switch which will permit to find the
     * ProxyManager which manages the proxy in a given context(thread)
     * @param pms
     */
    void jdoSetPersistenceManagerSwitch(ProxyManagerSwitch pms);


    /**
     * @return the PersistenceManagerFactory which represents the data support
     * inside which the proxy is persistent.
     */
    PersistenceManagerFactory jdoGetPersistenceManagerFactory();

    /**
     * It assignes a PersistenceManagerFactory
     * @param pmf is the PersistenceManagerFactory which represents the data
     * support inside which the proxy is persistent.
     */
    void jdoSetPersistenceManagerFactory(PersistenceManagerFactory pmf);

    /**
     * @return the TransactionalPersistenceManager which manage the
     * concurrency, loading, ... of the proxy.
     */
    TransactionalPersistenceManager jdoGetTransactionalPersistenceManager();

    /**
     * It assignes a TransactionalPersistenceManager
     * @param tpm is the TransactionalPersistenceManager which manage the
     * concurrency, loading, ... of the proxy.
     */
    void jdoSetTransactionalPersistenceManager(TransactionalPersistenceManager tpm);

    /**
     * It retrieves an object helping the configuration of the Jorm class.
     * Currently Jorm is not fractalize then the returned object is the PBinder
     * class name.
     */
    Object getJormConfig();

    void jdoMakePersistent(ProxyManager pm);

    void jdoDeletePersistent(ProxyManager pm);

    /**
     * @return return the mapper name by which the instance is persistent.
     */
    String jdoGetMapperName();

    /**
     * @return return the mapper name by which the instance is persistent.
     */
    String jdoGetProjectName();

    /**
     * @return The file name of the jdo descriptor. This file must be availlable
     * in the class at runtime.
     */
    String jdoGetJdoFileName();

    /**
     * Retrieves the hints needed to build the PName of the persistent object
     * during the export action.
     */
    Object jdoGePNameHints();

    /**
     * Assignes the hints needed to build the PName of the persistent object
     * during the export action.
     */
    void jdoSetPNameHints(Object hints);
}
