/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.generation.enhancer;

import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Constants;
import org.objectweb.asm.ClassReader;
import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.metadata.SpeedoField;
import org.objectweb.speedo.metadata.SpeedoModifier;
import org.objectweb.speedo.generation.api.SpeedoCompilerParameter;
import org.objectweb.util.monolog.api.Logger;

import java.io.ObjectStreamClass;
import java.util.Iterator;
import java.util.Map;
import java.util.Collection;

/**
 * Analyzes a class and updates the Speedo meta information accordingly.
 *
 * Adapted from storeClassInfo, verify and getField methods in EnhancerTool.
 */
public class ClassInfoAnalyzer extends LoggedClass implements ClassVisitor {

    /**
     * The access enhancer, used to load the super class(es) of the class, if
     * necessary.
     */
    final EnhancerComponent enhancer;

    /**
     * The Speedo meta information for the visited class.
     */
    final SpeedoClass jdoClass;

    /**
     * A collection of SpeedoXMLDescriptor describing known persistent classes
     */
    private final Collection xmlDescriptors;

    /**
     * Creates a new {@link ClassInfoAnalyzer}.
     *
     * @param enhancer the access enhancer, used to load the super class(es) of
     * the class, if necessary.
     * @param jdoClass the Speedo meta information for the visited class.
     */
    public ClassInfoAnalyzer(final EnhancerComponent enhancer,
                             final SpeedoClass jdoClass,
                             final Collection xmlDescriptors,
                             Logger logger) {
        super(logger);
        this.enhancer = enhancer;
        this.jdoClass = jdoClass;
        this.xmlDescriptors = xmlDescriptors;
    }

    // IMPLEMENTATION OF THE ClassVisitor INTERFACE //
    // ---------------------------------------------//

    public void visit(final int access,
                      final String name,
                      final String superName,
                      final String[] interfaces,
                      final String sourceFile) {
        // Stores class's signature
        jdoClass.signature = Util.modifier(access & ~Constants.ACC_SUPER);
        jdoClass.isAbstract = (access & Constants.ACC_ABSTRACT) != 0;

        // Verifies inheritance
        String superClass = jdoClass.persistenceCapableSuperClassName;
        if (superClass != null && !superClass.equals("") &&
                !superClass.equals(superName.replace('/', '.'))) {
            throw new RuntimeException("Class " + name + " has actually " +
                    superName + " as super class");
        }
        // Gets implemented interfaces
        for (int i = 0; i < interfaces.length; i++) {
            if (interfaces[i].equals("javax/jdo/InstanceCallbacks"))
                jdoClass.isInstanceCallbacks = true;
            if (interfaces[i].equals("java/io/Serializable"))
                jdoClass.isSerializable = true;
        }

        // Looks for fields defined in the jdoClass in the Java class and
        // assigns the corresponding modifier if missing
        ClassFieldFinder finder = new ClassFieldFinder();
        Iterator i = jdoClass.jdoField.values().iterator();
        //System.out.println("field names: " + jdoClass.jdoField.keySet());
        //System.out.println("fields [" + jdoClass.jdoField.values().size() + "]: " + jdoClass.jdoField.values());
        while (i.hasNext()) {
            SpeedoField current = ((SpeedoField) i.next());
            //System.out.println("treat  field " + current.name);
            SpeedoField f = finder.getField(current.name, name, jdoClass);
            if (f == null)
                throw new RuntimeException(
                        "Persistent Field " + current.name +
                        " not defined in class " + name);
            current.name = f.name;
            current.access = f.access;
            current.desc = f.desc;
            if (current.persistenceModifier == SpeedoModifier.missing)
                current.persistenceModifier =
                        (Util.isPersistentType(current.desc, xmlDescriptors)) ?
                        SpeedoModifier.persistent :
                        SpeedoModifier.none;
        }

        // Stores the serial version uid for serializable classes
        if (jdoClass.isSerializable) {
            try {
                ObjectStreamClass oserial = ObjectStreamClass.lookup(
                        Class.forName(name.replace('/', '.')));
                jdoClass.VersionUID = oserial.getSerialVersionUID();
            } catch (Exception e) {
            }
        }
    }

    public void visitInnerClass(final String name,
                                final String outerName,
                                final String innerName,
                                final int access) {
        // does nothing
    }

    public void visitField(final int access,
                           final String name,
                           final String desc,
                           final Object value) {
        // Adds in the object model fields whose type is known as
        // Persistence Capable
        if (!jdoClass.jdoField.containsKey(name)) {
            // This field is not defined in the jdoclass
            if (Util.isPersistentType(desc, xmlDescriptors)) {
                SpeedoField jdoField = new SpeedoField();
                jdoField.persistenceModifier = SpeedoModifier.persistent;
                jdoField.name = name;
                jdoField.access = access;
                jdoField.desc = desc;
                jdoClass.add(jdoField);
            }
        }
    }

    public CodeVisitor visitMethod(final int access,
                                   final String name,
                                   final String desc,
                                   final String[] exceptions) {
        // this visitor does not need to visit the code of methods
        return null;
    }

    public void visitEnd() {
        // Removes from the object model all fields that are not defined as
        // persistent
        Iterator i = jdoClass.jdoField.entrySet().iterator();
        while (i.hasNext()) {
            Map.Entry entry = (Map.Entry) i.next();
            SpeedoField field = (SpeedoField) entry.getValue();
            if (field.persistenceModifier != SpeedoModifier.persistent) {
                i.remove();
            }
        }
    }

    /**
     * Looks for a field in a class, and in its (persistent) super classes.
     */
    class ClassFieldFinder implements ClassVisitor {

        /**
         * The field which is looked for.
         */
        private String fieldName;

        /**
         * Information about this field, if found.
         */
        private SpeedoField result;

        public SpeedoField getField(final String fieldName,
                                    String className,
                                    SpeedoClass jdoClass) {
            this.fieldName = fieldName;
            SpeedoCompilerParameter scp = enhancer.getSpeedoCompilerParameter();

            try {
                while (true) {
                    // lookup in the class if the field exists
                    ClassReader cr;
                    if (scp.keepclasses) {
                        cr = enhancer.loadJavaClass(enhancer.isSrcJar,
                                className, scp.input, false, scp.output);
                    } else {
                        cr = enhancer.loadJavaClass(enhancer.isSrcJar,
                                className, scp.input, false);
                    }
                    cr.accept(this, true);
                    if (result != null) {
                        return result;
                    }

                    // lookup in the super class, if applicable
                    className = jdoClass.persistenceCapableSuperClassName;
                    if (className == null || className.length() == 0) {
                        // The current class has not a superclasses and the
                        // field has not been found
                        return null;
                    }

                    //fetch the SpeedoClass of the parent
                    jdoClass = Util.isPersistentCapable(className, xmlDescriptors);
                }
            } catch (Exception e) {
                e.printStackTrace();
                return null;
            }
        }

        public void visit(final int access,
                          final String name,
                          final String superName,
                          final String[] interfaces,
                          final String sourceFile) {
            // does nothing
        }

        public void visitInnerClass(final String name,
                                    final String outerName,
                                    final String innerName,
                                    final int access) {
            // does nothing
        }

        public void visitField(final int access,
                               final String name,
                               final String desc,
                               final Object value) {
            if (name.equals(fieldName)) {
                // we have found the field we are looking for
                result = new SpeedoField();
                result.name = name;
                result.access = access;
                result.desc = desc;
            }
        }

        public CodeVisitor visitMethod(final int access,
                                       final String name,
                                       final String desc,
                                       final String[] exceptions) {
            // this visitor does not need to visit the code of methods
            return null;
        }

        public void visitEnd() {
            // does nothing
        }
    }
}
