/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.generation.enhancer;

import org.objectweb.asm.ClassAdapter;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.CodeAdapter;
import org.objectweb.asm.Constants;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Set;

/**
 * Renames a class.
 *
 * Adapted from modifyClassName in EnhancerTool.
 */
public class ClassRenamer extends LoggedClassAdapter {

    /**
     * The old class name, and the names of its super classes.
     */
    private Set oldNames;

    /**
     * The new class name
     */
    private final String newName;

    /**
     * Constructs a new {@link ClassRenamer}.
     *
     * @param cv the visitor to be used to generate the renamed class
     * @param newName the new class name
     */
    public ClassRenamer(final ClassVisitor cv,
                        final Set oldNames,
                        final String newName,
                        Logger logger) {
        super(cv, logger);
        this.oldNames = oldNames;
        this.newName = newName;
    }

    // IMPLEMENTATION OF THE ClassVisitor INTERFACE //
    // ---------------------------------------------//

    public void visit(final int access,
                      final String name,
                      final String superName,
                      final String[] interfaces,
                      final String sourceFile) {
        oldNames.add(name);
        if (debug) {
            logger.log(BasicLevel.DEBUG,
                    "Rename the class '" + name + "' to '" + newName + "'");
        }
        cv.visit(access, newName, superName, interfaces, sourceFile);
    }

    public void visitInnerClass(final String name,
                                final String outerName,
                                final String innerName,
                                final int access) {
        cv.visitInnerClass(name, outerName, innerName, access); // TODO
    }

    public CodeVisitor visitMethod(final int access,
                                   final String name,
                                   final String desc,
                                   final String[] exceptions) {
        CodeVisitor cv = this.cv.visitMethod(access, name, desc, exceptions);
        if (cv != null && (access & Constants.ACC_ABSTRACT) == 0) {
            cv = new CodeRenamer(cv);
        }
        return cv;
    }

    /**
     * Modifies instructions that reference the old class.
     */
    class CodeRenamer extends CodeAdapter {

        /**
         * Creates a new {@link CodeRenamer}.
         *
         * @param cv the visitor to be used to generate the modified code
         */
        public CodeRenamer(final CodeVisitor cv) {
            super(cv);
        }

        // IMPLEMENTATION OF THE CodeVisitor INTERFACE //
        // --------------------------------------------//

        public void visitTypeInsn(int i, String s) {
            if (oldNames.contains(s)) {
                s = newName;
            }
            cv.visitTypeInsn(i, s);
        }

        public void visitFieldInsn(final int opcode,
                                   final String owner,
                                   final String name,
                                   final String desc) {
            if (oldNames.contains(owner)) {
                cv.visitFieldInsn(opcode, newName, name, desc);
            } else {
                cv.visitFieldInsn(opcode, owner, name, desc);
            }
        }

        public void visitMethodInsn(final int opcode,
                                    final String owner,
                                    final String name,
                                    final String desc) {
            if (oldNames.contains(owner)) {
                cv.visitMethodInsn(opcode, newName, name, desc);
            } else {
                cv.visitMethodInsn(opcode, owner, name, desc);
            }
        }
    }
}