/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.generation.enhancer;

import org.objectweb.asm.ClassAdapter;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Constants;
import org.objectweb.asm.Type;
import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.metadata.SpeedoField;
import org.objectweb.speedo.generation.lib.NamingRules;
import org.objectweb.speedo.api.UserId;
import org.objectweb.speedo.api.UserIdFactory;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Iterator;

/**
 * It adds the implementation of the UserIfFactory interface in a PName.
 *
 * @author S.Chassande-Barrioz
 */
public class PNameEnhancer extends LoggedClassAdapter {

    public final static String USER_ID_DESC =
            "L" + UserId.class.getName().replace('.', '/') + ";";
    public final static String PTYPE_CLASS_NAME =
            "org/objectweb/jorm/type/api/PType";
    public final static String PTYPE_DESC = "L" + PTYPE_CLASS_NAME + ";";

    protected SpeedoClass jdoClass;

    public PNameEnhancer(ClassVisitor classVisitor,
                          SpeedoClass jdoClass,
                         Logger logger) {
        super(classVisitor, logger);
        this.jdoClass = jdoClass;
    }

    public void visit(final int access,
                      final String className,
                      final String superName,
                      final String[] interfaces,
                      final String sourceFile) {
        String[] itfs;
        String uifcn = UserIdFactory.class.getName().replace('.', '/');
        boolean alreadyUIF = false;
        if (interfaces == null || interfaces.length==0) {
            itfs = new String[]{uifcn};
        } else {
            int i=0;
            while(i<interfaces.length && !interfaces[i].equals(uifcn)) {
                i++;
            }
            alreadyUIF = i<interfaces.length;
            if (alreadyUIF) {
                itfs = interfaces;
            } else {
                itfs = new String[interfaces.length + 1];
                System.arraycopy(interfaces, 0, itfs, 1, interfaces.length);
                itfs[0] = uifcn;
            }
        }
        if (alreadyUIF) {
            logger.log(BasicLevel.DEBUG, "The class " + className
                    + " already contains the interface " + uifcn);
            super.visit(access, className, superName, interfaces, sourceFile);
            return;
        }
        logger.log(BasicLevel.DEBUG, "Add to the class " + className
                + " the interface " + uifcn);
        super.visit(access, className, superName, itfs, sourceFile);

        //Add the method getNewUserId
        String uicn = jdoClass.objectidClass.replace('.', '/');
        String methodName = "getNewUserId";
        if (debug) {
            logger.log(BasicLevel.DEBUG, "Add the method '" + methodName
                    + "' in the class '" + className + "'");
        }
        logger.log(BasicLevel.DEBUG, "Add to the class " + className
                + " the method " + methodName);
        CodeVisitor cv = this.cv.visitMethod(
                Constants.ACC_PUBLIC,
                methodName,
                "()" + USER_ID_DESC,
                null);

        // GENERATE: Titi t = new Titi();
        // GENERATE: Titi is the class of the user identifier
        logger.log(BasicLevel.DEBUG, "add code: " + uicn + " t = new " + uicn + "();");
        cv.visitTypeInsn(Constants.NEW, uicn);
        cv.visitInsn(Constants.DUP);
        cv.visitMethodInsn(Constants.INVOKESPECIAL, uicn, "<init>", "()V");
        cv.visitVarInsn(Constants.ASTORE, 1);

        //t.f1 = this.f1
        boolean hasLongOrDouble = false;
        for(Iterator it = jdoClass.jdoField.values().iterator(); it.hasNext();) {
            SpeedoField sp = (SpeedoField) it.next();
            if (sp.primaryKey) {
                hasLongOrDouble |= sp.desc.charAt(0) == 'D'
                        || sp.desc.charAt(0) == 'J';
                logger.log(BasicLevel.DEBUG, "add code: this." + sp.name + " = t." + sp.name);
                cv.visitVarInsn(Constants.ALOAD, 1); //Can be a long|double value
                cv.visitVarInsn(Constants.ALOAD, 0);
                cv.visitFieldInsn(Constants.GETFIELD, className, sp.name, sp.desc);
                cv.visitFieldInsn(Constants.PUTFIELD, uicn, sp.name, sp.desc);
            }
        }

        // GENERATE: t.jdoSetPersistentClassName(this.getPType().getJormName());
        logger.log(BasicLevel.DEBUG, "add code: t.jdoSetPersistentClassName(this.getPType().getJormName());");
        cv.visitVarInsn(Constants.ALOAD, 1);
        cv.visitVarInsn(Constants.ALOAD, 0);
        cv.visitMethodInsn(Constants.INVOKEVIRTUAL,
                className, "getPType", "()" + PTYPE_DESC);
        cv.visitMethodInsn(Constants.INVOKEVIRTUAL,
                PTYPE_CLASS_NAME, "getJormName", "()Ljava/lang/String;");
        cv.visitMethodInsn(Constants.INVOKEVIRTUAL,
                uicn, "jdoSetPersistentClassName", "(Ljava/lang/String;)V");
        logger.log(BasicLevel.DEBUG, "add code: return t;");
        cv.visitVarInsn(Constants.ALOAD, 1);
        cv.visitInsn(Constants.ARETURN);
        cv.visitMaxs((hasLongOrDouble ? 3 : 2), 2); 
    }
}
