/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.generation.enhancer;

import org.objectweb.asm.ClassAdapter;
import org.objectweb.asm.ClassVisitor;
import org.objectweb.asm.CodeVisitor;
import org.objectweb.asm.Constants;
import org.objectweb.asm.Type;
import org.objectweb.speedo.metadata.SpeedoClass;
import org.objectweb.speedo.metadata.SpeedoField;
import org.objectweb.speedo.generation.lib.NamingRules;
import org.objectweb.speedo.api.UserId;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.Iterator;

/**
 * It adds the implementation of the UserId and specific PNameGetter interfaces
 * in the user identifier.
 *
 * @author S.Chassande-Barrioz
 */
public class UserIdEnhancer extends LoggedClassAdapter {

    protected SpeedoClass jdoClass;

    public final static String ADDED_FIELD_NAME = "jdoPersistentClassName";
    public final static String ADDED_FIELD_DESC = "Ljava/lang/String;";

    public UserIdEnhancer(ClassVisitor classVisitor,
                          SpeedoClass jdoClass,
                          Logger logger) {
        super(classVisitor, logger);
        this.jdoClass = jdoClass;
    }

    public void visit(final int access,
                      final String className,
                      final String superName,
                      final String[] interfaces,
                      final String sourceFile) {
        String[] itfs;
        String pngcn = NamingRules.pngName(jdoClass.objectidClass).replace('.', '/');
        boolean alreadyPNG = false;
        String sncn = UserId.class.getName().replace('.','/');
        if (interfaces == null || interfaces.length==0) {
            itfs = new String[]{pngcn, sncn};
        } else {
            int i=0;
            while(i<interfaces.length && !interfaces[i].equals(pngcn)) {
                i++;
            }
            alreadyPNG = i<interfaces.length;
            if (alreadyPNG) {
                itfs = interfaces;
            } else {
                itfs = new String[interfaces.length + 2];
                System.arraycopy(interfaces, 0, itfs, 2, interfaces.length);
                itfs[0] = pngcn;
                itfs[1] = UserId.class.getName().replace('.','/');
            }
        }
        if (alreadyPNG) {
            logger.log(BasicLevel.DEBUG, "The class " + className
                    + " already contains the interface " + pngcn + " and " + sncn);
            super.visit(access, className, superName, interfaces, sourceFile);
            return;
        }
        logger.log(BasicLevel.DEBUG, "Add to the class " + className
                + " the interface " + pngcn);
        logger.log(BasicLevel.DEBUG, "Add to the class " + className
                + " the interface " + sncn);
        super.visit(access, className, superName, itfs, sourceFile);

        for(Iterator it = jdoClass.jdoField.values().iterator(); it.hasNext();) {
            SpeedoField sp = (SpeedoField) it.next();
            if (sp.primaryKey) {
                String methodName = "pnGet" + upperFL(sp.name);
                logger.log(BasicLevel.DEBUG, "Add to the class " + className
                        + " the method " + methodName);
                CodeVisitor cv = this.cv.visitMethod(
                        Constants.ACC_PUBLIC,
                        methodName,
                        "(Ljava/lang/Object;)" + sp.desc,
                        null);
                cv.visitVarInsn(Constants.ALOAD, 0);
                cv.visitFieldInsn(Constants.GETFIELD, className, sp.name, sp.desc);
                Type returnType = Type.getType(sp.desc);
                cv.visitInsn(returnType.getOpcode(Constants.IRETURN));
                cv.visitMaxs(
                        (sp.desc.equals("J") || sp.desc.equals("D")) ? 2 : 1,
                        2);
            }
        }
        // declare the field jdoPersistentClassName
        //int access, String name, String desc, Object value
        logger.log(BasicLevel.DEBUG, "Add to the class " + className
                + " the field " + ADDED_FIELD_NAME);
        cv.visitField(Constants.ACC_PRIVATE, ADDED_FIELD_NAME,
                ADDED_FIELD_DESC, null);

        //Add the method jdoGetPersistentClassName
        String methodName = "jdoGetPersistentClassName";
        logger.log(BasicLevel.DEBUG, "Add to the class " + className
                + " the method " + methodName);
        CodeVisitor cv = this.cv.visitMethod(
                Constants.ACC_PUBLIC,
                methodName,
                "()Ljava/lang/String;",
                null);
        cv.visitVarInsn(Constants.ALOAD, 0);
        cv.visitFieldInsn(Constants.GETFIELD,
                className, ADDED_FIELD_NAME, ADDED_FIELD_DESC);
        Type returnType = Type.getType(ADDED_FIELD_DESC);
        cv.visitInsn(returnType.getOpcode(Constants.IRETURN));
        cv.visitMaxs(1, 2);

        //Add the method jdoSetPersistentClassName
        methodName = "jdoSetPersistentClassName";
        logger.log(BasicLevel.DEBUG, "Add to the class " + className
                + " the method " + methodName);
        cv = this.cv.visitMethod(
                Constants.ACC_PUBLIC,
                methodName,
                "(Ljava/lang/String;)V",
                null);
        cv.visitVarInsn(Constants.ALOAD, 0);
        cv.visitVarInsn(Constants.ALOAD, 1);
        cv.visitFieldInsn(Constants.PUTFIELD,
                className, ADDED_FIELD_NAME, ADDED_FIELD_DESC);
        cv.visitInsn(Constants.RETURN);
        cv.visitMaxs(2, 2);
    }

    private String upperFL(String str) {
        if (str == null || str.length()==0) {
            return str;
        } else if (str.length() == 1) {
            return String.valueOf(Character.toUpperCase(str.charAt(0)));
        } else {
            return Character.toUpperCase(str.charAt(0)) + str.substring(1);
        }
    }
}
