/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 0.1
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: P. Dechamboux, Y. Devaux, F. Dhumeaux.
 *
 */

package org.objectweb.speedo.generation.lib;

import org.objectweb.speedo.metadata.SpeedoClass;

import java.util.StringTokenizer;

/**
 * Normalises names given to Proxy infrastructure's files created during
 * generation.
 * @author	Y. Devaux
 */
public abstract class NamingRules {

    /**
     * Suffix added to the class name to create the proxy class name.
     */
    public final static String proxy = "Proxy";

    /**
     * Suffix added to the class name to create the field manager class name.
     */
    public final static String fields = "Fields";

    /**
     * Suffix added to the object idclass name
     */
    public final static String objectId = "_Id";

    /**
     * Gives the proxy name.
     *
     * @param className class name.
     * @return proxy name.
     */
    public final static String proxyName(String className) {
        return className + proxy;
    }

    public final static String generatedObjectIdName(String className) {
        return className + objectId;
    }

    /**
     * Gives the name of the persistent fields holder.
     *
     * @param className className.
     * @return name for persistent fields holder.
     */
    public final static String fieldsName(String className) {
        return className + fields;
    }

    /**
     * Gives the name for the JORM class which manages binding between memory
     * instances and database instances.
     *
     * @param className class name.
     * @return binder name.
     */
    public final static String bindingName(String className) {
        return className + "Binding";
    }

    /**
     * Gives the name for the JORM class which manages binding between memory
     * instances and database instances.
     *
     * @param className class name.
     * @return binder name.
     */
    public final static String fqBindingName(String className, String mapperName) {
        String p = packageName(className);
        if (p == null || p.length() == 0) {
            return mapperName + "." + className + "Binding";
        } else {
            return p + "." + mapperName + "." + className(className) + "Binding";
        }
    }

    /**
     * Gives the name for the JORM class which manages mapping.
     *
     * @param className class name.
     * @return mapping name.
     */
    public final static String mappingName(String className) {
        return className + "Mapping";
    }

    /**
     * Gives the name for the JORM class which manages mapping.
     *
     * @param compositeNameName composite name.
     * @return mapping name.
     */
    public final static String binderName(String compositeNameName) {
        return compositeNameName + "Binder";
    }

    /**
     * Gives the name for the JORM class which manages mapping.
     *
     * @param compositeNameName composite name.
     * @return mapping name.
     */
    public final static String pnameName(String compositeNameName) {
        return compositeNameName + "PName";
    }

    /**
     * Gives the name for the JORM class which manages mapping.
     *
     * @param compositeNameName composite name.
     * @return mapping name.
     */
    public final static String pngName(String compositeNameName) {
        return compositeNameName + "PNG";
    }

    /**
     * Gives the name for the JORM class which manages mapping.
     *
     * @param className class name.
     * @return mapping name.
     */
    public final static String fqMappingName(String className, String mapperName) {
        String p = packageName(className);
        if (p == null || p.length() == 0) {
            return mapperName + "." + className + "Mapping";
        } else {
            return p + "." + mapperName + "." + className(className) + "Mapping";
        }
    }

    /**
     * Gives the name for the object which implements the interface PAccessor.
     *
     * @param className class name.
     * @return manager name.
     */
    public final static String accessorName(String className) {
        return className + "Accessor";
    }

    /**
     * Gives the name for key objects.
     *
     * @param className class name.
     * @return key class name.
     */
    public final static String keyName(String className) {
        return className + "Key";
    }

    /**
     * Gives the name of an accessor for a field (a get method).
     *
     * @param fieldName field name
     * @return the name of the accessor method.
     */
    public final static String getterName(SpeedoClass sc, String fieldName) {
        return "jdoGet" + fieldName;
    }

    /**
     * Gives the name of a mutator for a field (a set method).
     *
     * @param fieldName field name
     * @return the name of the mutator method.
     */
    public final static String setterName(SpeedoClass sc, String fieldName) {
        return "jdoSet" + fieldName;
    }

    /**
     * Gets the name of a class
     *
     * @param completeClassName the complete name of the class including its package name
     * @return the name of the class without its package
     */
    public static String className(String completeClassName) {
        StringTokenizer tok = new StringTokenizer(completeClassName, ".");
        String res = "";
        while (tok.hasMoreTokens()) res = tok.nextToken();
        return res;
    }

    /**
     * Gets the package of a class
     *
     * @param completeClassName the complete name of the class including its package name
     * @return the package of the class without its class name
     */
    public static String packageName(String completeClassName) {
        StringTokenizer tok = new StringTokenizer(completeClassName, ".");
        String res = "";
        String resold = "";
        while (tok.hasMoreTokens()) {
            resold = (res == "") ? "" : ((resold == "") ? res : resold + "." + res);
            res = tok.nextToken();
        }
        return resold;
    }
}
