/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.runtime;

import org.objectweb.fractal.api.control.UserBindingController;
import org.objectweb.jorm.api.PClassMapping;
import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.api.PClassMappingCtrl;
import org.objectweb.jorm.api.PMapCluster;
import org.objectweb.jorm.facility.naming.polymorphid.PolymorphIdMgrImpl;
import org.objectweb.jorm.facility.naming.polymorphid.PolymorphIdPName;
import org.objectweb.jorm.facility.naming.basidir.BasidBinder;
import org.objectweb.jorm.metainfo.api.Package;
import org.objectweb.jorm.metainfo.lib.BasicClass;
import org.objectweb.jorm.metainfo.lib.BasicCompositeName;
import org.objectweb.jorm.metainfo.lib.JormManager;
import org.objectweb.jorm.naming.api.PBinder;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.jorm.naming.api.PNameManager;
import org.objectweb.jorm.naming.api.PNamingContext;
import org.objectweb.jorm.naming.api.PNameCoder;
import org.objectweb.jorm.runtime.JormPathHelper;
import org.objectweb.speedo.api.JormFactory;
import org.objectweb.speedo.api.SpeedoProxy;
import org.objectweb.speedo.api.SpeedoGenClassProxy;
import org.objectweb.speedo.api.UserId;
import org.objectweb.speedo.api.UserIdFactory;
import org.objectweb.speedo.api.ExceptionHelper;
import org.objectweb.speedo.generation.lib.NamingRules;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.perseus.cache.api.CacheManager;

import javax.jdo.JDOException;
import javax.jdo.JDOFatalInternalException;
import java.io.File;
import java.io.ObjectInputStream;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.List;
import java.util.ArrayList;

/**
 * This class manages both naming sorts used in Speedo (Container id and user
 * id). The container id (long, long) management is delegated to a polymorphId
 * manager (PolymorphIdMgrImpl) whereas the user id is directly managed by this
 * class.
 * In JDO the binder instance can be used also as PNamingContext because all
 * classes in a class hierarchy must have the same identifier strucutre.
 * Then it is not necessary to use intermediates PNamingContext. Then a PName is
 * always valid inside a PBinder.
 *
 * @author S.Chassande-Barrioz
 */
public class BasicJormFactory
        implements JormFactory, UserBindingController {

    public final static String MAPPER_BINDING = "mapper";
    public final static String LOGGER_FACTORY_BINDING = "logger-factory";
    public final static String CACHE_MANAGER_BINDING = "cache-manager";
    public final static String LOGGER_NAME
            = "org.objectweb.speedo.rt.jormFactory";

    public final static String CID_BINDER_CLASS_NAME
            = "org.objectweb.jorm.facility.naming.polymorphid.PolymorphIdBinderInfo";
    public final static String CID_PNM_CLASS_NAME
            = "org.objectweb.jorm.facility.naming.polymorphid.PolymorphRefNC";

    public final static String BASID_BINDER_CLASS_NAME
            = "org.objectweb.jorm.facility.naming.basidir.BasidBinder";

    public final static boolean CODING_ONLY = false;

    /**
     * The mapper is used to map and find the Jorm classes.
     */
    protected PMapper mapper = null;

    protected LoggerFactory loggerFactory = null;

    protected CacheManager cache = null;

    /**
     * The manager of the naming used by the container (long, long). It
     * provides PBinder, PNamingContext since a class name.
     */
    protected PolymorphIdMgrImpl cIdManager = null;

    /**
     * It contains the association between a class name and its binder. This map
     * contains only the binders which manage user pname structures.
     */
    protected Map uIdBinders = null;
    protected Logger logger = null;

    /**
     * builds the BasicJormFactory and instanciate the polymorphid manager
     * and the hosting structure for the binder referencemet.
     */
    public BasicJormFactory() {
        cIdManager = null;
        uIdBinders = new HashMap();
    }

    public PolymorphIdMgrImpl getcIdManager() {
        if (cIdManager == null) {
            cIdManager = new PolymorphIdMgrImpl();
            cIdManager.setLogger(logger);
            try {
                cIdManager.init(mapper, PClassMapping.CREATE_STRUCTURE_IF_NEEDED);
            } catch (PException e) {
                Exception ie = ExceptionHelper.getNested(e);
                logger.log(BasicLevel.ERROR,
                        "Impossible to initialize the container id manager ",
                        ie);
                throw new RuntimeException(
                        "Impossible to initialize the container id manager "
                        + ie.getMessage());
            }
        }
        return cIdManager;
    }

    // IMPLEMENTATION OF THE UserBindingController INTERFACE //
    //-------------------------------------------------------//

    public Object getFcBindings(String s) {
        if (LOGGER_FACTORY_BINDING.equals(s))
            return loggerFactory;
        else if (MAPPER_BINDING.equals(s))
            return mapper;
        else if (CACHE_MANAGER_BINDING.equals(s))
            return cache;
        else
            return null;
    }

    public void addFcBinding(String s, Object o) {
        if (LOGGER_FACTORY_BINDING.equals(s)) {
            loggerFactory = (LoggerFactory) o;
            if (loggerFactory != null) {
                logger = loggerFactory.getLogger(LOGGER_NAME);
            }
        } else if (MAPPER_BINDING.equals(s)) {
            mapper = (PMapper) o;
        } else if (CACHE_MANAGER_BINDING.equals(s)) {
            cache = (CacheManager) o;
        }
    }

    public void removeFcBinding(String s, Object serverItf) {
        if (LOGGER_FACTORY_BINDING.equals(s)) {
            loggerFactory = null;
        } else if (MAPPER_BINDING.equals(s)) {
            mapper = null;
        } else if (CACHE_MANAGER_BINDING.equals(s)) {
            cache = null;
        }
    }


    // IMPLEMENTATION OF THE JormFactory INTERFACE //
    //---------------------------------------------//

    /**
     * @return the PBinder instance which manages the identifer of the
     * SpeedoProxy instance
     */
    public PBinder getPBinder(SpeedoProxy sp) throws PException {
        String clName = sp.getClass().getName();
        synchronized (mapper) {
            PClassMapping pcm = mapper.lookup(clName);
            if (pcm != null)
                return pcm.getPBinder();
        }
        return getPClassMapping(sp).getPBinder();
    }

    /**
     * This Method is a shortcut to the getPBinder(SpeedoProxy) method. It only
     * does the instanciation of the classname
     * (Class.ForName(classsName).newInstance()) and call the
     * getPBinder(SpeedoProxy) method with the created instance. Then if an
     * instance is availlable it is better to use the other method.
     * @param classname the Jorm class name managed by the wanted PBinder
     * @return the PBinder instance to use for the given jorm class name
     * @throws PException
     */
    public PBinder getPBinder(String classname) throws PException {
        return getPBinder(getInstance(classname));
    }

    /**
     * Retrieves a PClassMapping instance managing a SpeedoProxy. If the
     * PClassMapping does not already exist then it is created, configured and
     * mapped in the PMapper. The PClassMapping configuration contains the
     * PBinder assignement, the PNamingContext assignement for each reference,
     * and the GenClassMapping assignement. The assigned GenClassMapping must
     * also be configured. In other hand the PClassMapping of referenced class
     * is not configured now. Only the PNamingContext is needed.
     * @param sp is the SpeedoProxy instance which the PClassMapping is required
     * @return the PClassMapping which manages a SpeedoProxy
     * @throws PException
     */
    public PClassMapping getPClassMapping(SpeedoProxy sp) throws PException {
        if (sp instanceof SpeedoGenClassProxy) {
            // The GenClassMapping must be already created before
            SpeedoGenClassProxy sgcp = (SpeedoGenClassProxy) sp;
            logger.log(BasicLevel.DEBUG,
                    "Looking for the PClassMapping of the generic class "
                    + sgcp.jdoGetGenClassId());
            return JormPathHelper.getPClassMapping(sgcp.jdoGetGenClassId(), mapper);
        }
        String clName = sp.getClass().getName();
        PClassMapping pcm = mapper.lookup(clName);
        if (pcm == null) {
            synchronized (mapper) {
                pcm = mapper.lookup(clName);
                if (pcm == null) {
                    // Creates the PClassMapping, its PBinder and the PNC
                    logger.log(BasicLevel.DEBUG,
                            "Looking for the PClassMapping of the class " + clName);
                    String pcmcn = NamingRules.fqMappingName(
                            clName, sp.jdoGetMapperName());
                    try {
                        pcm = (PClassMapping) java.lang.Class.forName(pcmcn).newInstance();
                    } catch (Exception e) {
                        throw new PException(e,
                                "Impossible to instanciate the PClassMapping of the class "
                                + clName + ": " + pcmcn);
                    }
                    logger.log(BasicLevel.DEBUG, "Jorm config:");
                    String[][] jormConf = (String[][]) sp.getJormConfig();
                    Map cn2bcn = new HashMap(jormConf.length);
                    for (int i = 0; i < jormConf.length; i++) {
                        logger.log(BasicLevel.DEBUG, jormConf[i][0] + " ==> " + jormConf[i][1]);
                        cn2bcn.put(jormConf[i][0], jormConf[i][1]);
                    }

                    // Fetch the Binder by the DataStoreNaming
                    PBinder binder = findPBinder(
                            (String) cn2bcn.get(JormPathHelper.getPath(clName)), clName);
                    pcm.setPBinder(binder);
                    binder.setPClassMapping(pcm);
                    //Assign the PNamingContext of the reference
                    RefConfig refConfig = new RefConfig(cn2bcn, sp.jdoGetMapperName());
                    pcm.configureRefFields(refConfig);
                    if (logger.isLoggable(BasicLevel.DEBUG)) {
                        logger.log(BasicLevel.DEBUG, "PClassMapping/PBinder/PNC created for the class " + clName);
                        logger.log(BasicLevel.DEBUG, "- pcm: " + pcm);
                        logger.log(BasicLevel.DEBUG, "- binder: " + binder);
                    }

                    // do the 'Map' operation on the mapper and register the pcm
                    // in the local map
                    mapper.map(null, pcm);
                    PMapCluster cluster =  mapper.getPMappingStructuresManager().getPMapCluster(clName);
                    if (cluster.isDefined()) {
                        cluster.createMappingStructures(false);
                    } else {
                        getPClassMapping((String)cluster.getUnResolvedDependencies().iterator().next());
                    }

                    if (logger.isLoggable(BasicLevel.DEBUG)) {
                        logger.log(BasicLevel.DEBUG, "Class " + clName + " mapped");
                    }
                    loadJormMI(sp, sp.getClass().getClassLoader());
                } else if (logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG, "PClassMapping for class " + clName
                            + " already exist:" + pcm);
            }
        }
        return pcm;
    }

    /**
     * This Method is a shortcut to the getPClassMapping(SpeedoProxy) method. It only
     * does the instanciation of the classname
     * (Class.ForName(classsName).newInstance()) and call the
     * getPClassMapping(SpeedoProxy) method with the created instance. Then if an
     * instance is availlable it is better to use the other method.
     *
     * Retrieves a PClassMapping instance managing a jorm class. If the
     * PClassMapping does not already exist then it is created, configured and
     * mapped in the PMapper. The PClassMapping configuration contains the
     * PBinder assignement, the PNamingContext assignement for each reference,
     * and the GenClassMapping assignement. The assigned GenClassMapping must
     * also be configured. In other hand the PClassMapping of referenced class
     * is not configured now. Only the PNamingContext is needed.
     * @param classname is the class name of the jorm class which the
     * PClassMapping is required
     * @return the PClassMapping which manages a SpeedoProxy
     * @throws PException
     */
    public PClassMapping getPClassMapping(String classname) throws PException {
        synchronized (mapper) {
            PClassMapping pcm = mapper.lookup(classname);
            if (pcm == null)
                pcm = getPClassMapping(this.getInstance(classname));
            return pcm;
        }
    }

    /**
     * @return the PNamingContext instance which manages a reference to the
     * SpeedoProxy instance
     */
    public PNamingContext getPNamingContext(SpeedoProxy sp) throws PException {
        if (sp.jdoIsContainerId())
            return getcIdManager().getRefNC(sp.getClass().getName());
        else
            return (PNamingContext) getPBinder(sp);
    }

    /**
     * This Method is a shortcut to the getPNamingContext(SpeedoProxy) method.
     * It only does the instanciation of the classname
     * (Class.ForName(classsName).newInstance()) and call the
     * getPNamingContext(SpeedoProxy) method with the created instance. Then if
     * an instance is availlable it is better to use the other method.
     * @param classname the Jorm class name managed by the wanted PNamingContext
     * @return the PNamingContext instance to use for the given jorm class name
     * @throws PException
     */
    public PNamingContext getPNamingContext(String classname) throws PException {
        return getPNamingContext(getInstance(classname));
    }

    /**
     * It retrieves a PName since a representation of an identifier. The
     * representation can be a PName which the PNamingContext reference must be
     * checked and updated if needed. The identifier representation can be also
     * an encode version of a PName. The coding format depends on the
     * implementation.
     *
     * @param oid is the identifier representation
     * @return a PName matching the oid representation and valid in a
     * PNameManager
     * @throws PException if the oid does not represent a managed identifier
     * representation.
     */
    public PName decode(Object oid) throws PException {
        if (oid instanceof PolymorphIdPName) {
            PolymorphIdPName pn = (PolymorphIdPName) oid;
            PNameManager pnm = null;
            if (pn.getPNameManager() instanceof PNamingContext) {
                pnm = getcIdManager().getPBinder(pn.pnGetClassId(null));
            } else {
                pnm = getcIdManager().getPBinder(pn.pnGetClassId(null));
            }
            if (pnm == pn.getPNameManager())
                pn.setPNamingContext((PNamingContext) pnm);
            return pn;
        } else if (oid instanceof UserId) {
            PBinder binder = getPBinder(
                    ((UserId) oid).jdoGetPersistentClassName());
            return binder.decodeAbstract(oid, null);
        } else if (oid instanceof PName) {
            return (PName) oid;
        } else if (oid instanceof String) {
            String stroid = (String) oid;
            int idx = stroid.indexOf("!");
            if (idx == -1) {
                throw new PException("Umanaged identifier: " + oid);
            }
            String fqcn = stroid.substring(0, idx);
            return getPBinder(fqcn).decodeString(stroid.substring(idx + 1));
        } else
            throw new PException("Umanaged identifier: " + oid);
    }

    /**
     * It retrieves a representation of an identifier since a PName.
     *
     * @param pn is the pname to encode
     * @return an identifier representation
     * @throws PException if the pname is malformed (null for example).
     */
    public Object encode(PName pn) throws PException {
        if (pn instanceof UserIdFactory) {
            return ((UserIdFactory) pn).getNewUserId();
        } else {
            return pn.getPNameManager().getPType().getJormName()
                + "!" +  pn.encodeString();
        }
    }
    // PRIVATE METHODS //
    //-----------------//

    private SpeedoProxy getInstance(String classname) throws PException {
        try {
            return (SpeedoProxy)
                    java.lang.Class.forName(classname).newInstance();
        } catch (Exception e) {
            throw new PException(e, "Impossible to instanciate the class "
                    + classname);
        }

    }

    private PBinder findPBinder(String bcn, String clName) throws PException {
        if (bcn == null || bcn.length() == 0)
            throw new JDOFatalInternalException(
                    "Impossible to get the PNamingContext of the class " + clName
                    + ": Specified binder class is not valid: " + bcn);
        if (bcn.equals(CID_BINDER_CLASS_NAME)) {
            return getcIdManager().getPBinder(clName);
        } else if (bcn.startsWith(BASID_BINDER_CLASS_NAME)) {
            short ct = Short.parseShort(
                bcn.substring(BASID_BINDER_CLASS_NAME.length()));
            BasidBinder binder = new BasidBinder(ct);
            uIdBinders.put(clName, binder);
            return binder;
        } else {
            synchronized (uIdBinders) {
                PBinder binder = (PBinder) uIdBinders.get(clName);
                if (binder != null)
                    return binder;
                try {
                    binder = (PBinder)
                            java.lang.Class.forName(bcn).newInstance();
                } catch (Exception e) {
                    throw new PException(e,
                            "Impossible to instanciate the binder of the class "
                            + clName + ": " + bcn);
                }
                binder.setCacheManager(cache);
                uIdBinders.put(clName, binder);
                return binder;
            }
        }
    }

    protected PNameManager findPNameManager(String pnmcn, String clName) throws PException {
        if (pnmcn == null || pnmcn.length() == 0)
            throw new JDOFatalInternalException(
                    "Impossible to get the PNamingContext of the class " + clName
                    + ": Specified PNameManager class is not valid: " + pnmcn);
        if (pnmcn.equals(CID_BINDER_CLASS_NAME)) {
            return getcIdManager().getRefNC(clName);
        } else if (pnmcn.startsWith(BASID_BINDER_CLASS_NAME)) {
            short ct = Short.parseShort(
                pnmcn.substring(BASID_BINDER_CLASS_NAME.length()));
            BasidBinder binder = new BasidBinder(ct);
            uIdBinders.put(clName, binder);
            return binder;
        } else {
            synchronized (uIdBinders) {
                PBinder binder = (PBinder) uIdBinders.get(clName);
                if (binder != null)
                    return binder;
                try {
                    binder = (PBinder)
                            java.lang.Class.forName(pnmcn).newInstance();
                } catch (Exception e) {
                    throw new PException(e,
                            "Impossible to instanciate the PNameManager of the class "
                            + clName + ": " + pnmcn);
                }
                uIdBinders.put(clName, binder);
                return binder;
            }
        }
    }

    private class RefConfig implements PClassMapping.ReferenceConfigurator {

        Map cn2bcn = null;
        String mapperName = null;
        List gcms = null;

        public RefConfig(Map cn2bcn, String mapperName) {
            this.cn2bcn = cn2bcn;
            this.mapperName = mapperName;
            this.gcms = new ArrayList();
        }

        public List getGenClassMappings() {
            return gcms;
        }

        // IMPLEMENTATION OF THE PClassMapping.PNCGetter INTERFACE //
        //---------------------------------------------------------//


        public PNameCoder getPNameCoder(String sourceclassName,
                                              String refFieldName,
                                              String destclassName) {
            try {
                PNamingContext pnc = getPNamingContext(destclassName);
                if (logger.isLoggable(BasicLevel.DEBUG))
                    logger.log(BasicLevel.DEBUG,
                            "Assign the PNamingContext for the field " + sourceclassName
                            + "." + refFieldName + " which references the class "
                            + destclassName + ": " + pnc);
                return pnc;
            } catch (PException e) {
                throw new JDOFatalInternalException(
                        "ERROR during the assignation of the PNameManager of the reference "
                        + refFieldName + " of the class " + sourceclassName, e);
            }
        }

        public PNameCoder getPNameCoder(String sourceclassName,
                                              String refFieldName,
                                              String[] genClassNames) {
            String cn = JormPathHelper.getPath(
                    sourceclassName, refFieldName, genClassNames);
            String pnmcn = (String) cn2bcn.get(cn);
            logger.log(BasicLevel.DEBUG,
                    "getPNameManager(" + cn + ") " + pnmcn);
            try {
                return (PNamingContext) findPNameManager(pnmcn, cn);
            } catch (PException e) {
                throw new JDOFatalInternalException(e.getMessage(), e);
            }
        }

        public PClassMapping getGenClassMapping(String sourceclassName,
                                                String refFieldName,
                                                String[] genClassNames) {

            String cn = JormPathHelper.getPath(
                    sourceclassName, refFieldName, genClassNames);
            String pnmcn = (String) cn2bcn.get(cn);
            logger.log(BasicLevel.DEBUG,
                    "getGenClassMapping(" + cn + ") pnmcn=" + pnmcn);
            try {
                PClassMapping gcm = (PClassMapping)
                        Class.forName(getGCMClassName(mapperName)).newInstance();
                PBinder pb = findPBinder(pnmcn, cn);
                gcm.setPBinder(pb);
                pb.setPClassMapping(gcm);
                gcms.add(gcm);
                return gcm;
            } catch (Exception e) {
                throw new JDOFatalInternalException(e.getMessage(), e);
            }
        }

        public PClassMapping getGenClassMapping(String sourceclassName,
                                                String refFieldName,
                                                String[] genClassNames,
                                                String destclassName) {
            String cn = JormPathHelper.getPath(
                    sourceclassName, refFieldName, genClassNames);
            String pnmcn = (String) cn2bcn.get(cn);
            logger.log(BasicLevel.DEBUG,
                    "getGenClassMapping(" + cn + ") pnmcn=" + pnmcn);
            try {
                PClassMapping gcm = (PClassMapping)
                        Class.forName(getGCMClassName(mapperName)).newInstance();
                PBinder pb = findPBinder(pnmcn, cn);
                gcm.setPBinder(pb);
                pb.setPClassMapping(gcm);
                PNamingContext pnc = getPNamingContext(destclassName);
                logger.log(BasicLevel.DEBUG,
                        "\tgetGenClassMapping: " + cn + "  pnc=" + pnc);
                ((PClassMappingCtrl) gcm).setPNameCoder(pnc);
                gcms.add(gcm);
                return gcm;
            } catch (Exception e) {
                throw new JDOFatalInternalException(e.getMessage(), e);
            }
        }
    }

    private void loadJormMI(SpeedoProxy sp, ClassLoader cl) {
        String clName = sp.getClass().getName();
        JormManager m = (JormManager) mapper.getMetaInfoManager();

        if (m.getClass(clName) != null) {
            return;
        }
        Set mos = null;
        try {
            String fn = sp.jdoGetJdoFileName();
            fn = fn.substring(0, fn.length() - 4);
            fn = fn.replace('.', File.separatorChar);
            fn += ".jmi";
            logger.log(BasicLevel.DEBUG,
                    "Loading the file " + fn + " from the classpath");
            ObjectInputStream ois = new ObjectInputStream(cl.getResourceAsStream(fn));
            mos = (Set) ois.readObject();
            ois.close();
        } catch (Exception e) {
            throw new JDOException("Impossible to load the jorm meta " +
                    "information for the class " + clName, e);
        }

        for (Iterator it = mos.iterator(); it.hasNext();) {
            Object o = it.next();
            if (o instanceof BasicClass) {
                BasicClass c = (BasicClass) o;
                Package s = m.createPackage(((Package) c.getParent()).getName());
                s.addClass(c);
                c.setParent(s);
                c.setLogger(m.getLogger());
                logger.log(BasicLevel.DEBUG,
                        "Jorm Meta Object Class " + c.getFQName() + " loaded");
            } else if (o instanceof BasicCompositeName) {
                BasicCompositeName cn = (BasicCompositeName) o;
                Package s = m.createPackage(((Package) cn.getParent()).getName());
                s.addCompositeName(cn);
                cn.setLogger(m.getLogger());
                logger.log(BasicLevel.DEBUG,
                        "Jorm Meta Object CompositeName " + cn.getFQName() + " loaded");
            } else
                throw new JDOException("Umanaged Jorm Meta Object " + o);
        }
    }

    private String getGCMClassName(String mapperName) {
        if (mapperName.startsWith("rdb")) {
            return "org.objectweb.jorm.interp.genclass.rdb.lib.RdbGenClassMapping";
        } else {
            throw new JDOFatalInternalException("Umanaged mapper: " + mapperName);
        }
    }
}
