/**
 * copyright area
 */
package org.objectweb.speedo.runtime;

import org.objectweb.fractal.api.ComponentIdentity;
import org.objectweb.fractal.api.InterfaceReference;
import org.objectweb.fractal.api.control.UserBindingController;
import org.objectweb.fractal.julia.control.content.SubComponentController;
import org.objectweb.fractal.util.Fractal;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.runtime.Mapper;
import org.objectweb.perseus.pool.api.Pool;
import org.objectweb.perseus.pool.api.PoolMatchFactory;
import org.objectweb.perseus.pool.api.PoolProvider;
import org.objectweb.perseus.pool.api.PoolResource;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.speedo.api.ProxyManager;
import org.objectweb.speedo.api.ProxyManagerSwitch;
import org.objectweb.speedo.runtime.mapper.JCAMapper;
import org.objectweb.speedo.runtime.mapper.JDBCMapper;
import org.objectweb.speedo.runtime.mapper.ConnectionSpecFactory;
import org.objectweb.speedo.Speedo;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOUnsupportedOptionException;
import javax.jdo.PersistenceManager;
import javax.jdo.PersistenceManagerFactory;
import javax.resource.cci.ConnectionSpec;
import java.util.Collection;
import java.util.Properties;

/**
 * @author S.Chassande-Barrioz
 */
public class SpeedoPersistenceManagerFactory
        implements PersistenceManagerFactory,
        PoolProvider,
        PoolMatchFactory,
        UserBindingController {

    public final static String PERSISTENCE_MANAGER_TEMPLATE_BINDING = "persistence-manager-template";
    public final static String MAPPER_BINDING = "mapper";
    public final static String PM_POOL_BINDING = "persistence-manager-pool";
    public final static String PROXY_MANAGER_SWITCH_BINDING = "proxy-manager-switch";
    public final static String LOGGER_FACTORY_BINDING = "logger-factory";
    public final static String COMPONENT_IDENTITY_BINDING = "component-identity";
    public final static String LOGGER_NAME
            = "org.objectweb.speedo.rt.SpeedoPersistenceManagerFactory";

    /**
     * Pool of PersistenceManager
     */
    private Pool managedPM = null;

    /**
     * Provides Loggers to all of the JDO implementation
     */
    private LoggerFactory loggerFactory = null;

    /**
     * Manages the log's traces
     */
    private Logger logger = null;

    /**
     * JORM object encapsulating the datastore
     */
    private PMapper mapper = null;

    private Properties connectionProperties = null;

    private ComponentIdentity pmTemplate;

    private ProxyManagerSwitch pms = null;

    private Object _this = null;

    private ConnectionSpecFactory csf = null;

    public SpeedoPersistenceManagerFactory() {
        connectionProperties = new Properties();
    }

    // IMPLEMENTATION OF THE UserBindingController INTERFACE //
    //-------------------------------------------------------//

    public Object getFcBindings(String s) {
        if (LOGGER_FACTORY_BINDING.equals(s))
            return loggerFactory;
        else if (PM_POOL_BINDING.equals(s))
            return managedPM;
        else if (PERSISTENCE_MANAGER_TEMPLATE_BINDING.equals(s))
            return pmTemplate;
        else if (MAPPER_BINDING.equals(s))
            return mapper;
        else if (PROXY_MANAGER_SWITCH_BINDING.equals(s))
            return pms;
        else
            return null;
    }

    public void addFcBinding(String s, Object o) {
        if (LOGGER_FACTORY_BINDING.equals(s)) {
            loggerFactory = (LoggerFactory) o;
            if (loggerFactory != null)
                logger = loggerFactory.getLogger(LOGGER_NAME);
        } else if (PM_POOL_BINDING.equals(s)) {
            managedPM = (Pool) o;
        } else if (PERSISTENCE_MANAGER_TEMPLATE_BINDING.equals(s)) {
            pmTemplate = (ComponentIdentity) o;
        } else if (MAPPER_BINDING.equals(s)) {
            mapper = (PMapper) o;
        } else if (PROXY_MANAGER_SWITCH_BINDING.equals(s)) {
            pms = (ProxyManagerSwitch) o;
        } else if (COMPONENT_IDENTITY_BINDING.equals(s)) {
            _this = o;
        }
    }

    public void removeFcBinding(String s, Object serverItf) {
        if (LOGGER_FACTORY_BINDING.equals(s)) {
            loggerFactory = null;
        } else if (PM_POOL_BINDING.equals(s)) {
            managedPM = null;
        } else if (PERSISTENCE_MANAGER_TEMPLATE_BINDING.equals(s)) {
            pmTemplate = null;
        } else if (MAPPER_BINDING.equals(s)) {
            mapper = null;
        } else if (PROXY_MANAGER_SWITCH_BINDING.equals(s)) {
            pms = null;
        }
    }

    // IMPLEMENTATION OF THE PersistentManagerFactory INTERFACE //
    //----------------------------------------------------------//

    /** Get an instance of PersistenceManager from this factory. The instance has
     * default values for options.
     * Invokes <code>init</code> at the first call.
     *
     * @return a PersistenceManager instance with default options.
     */
    public synchronized PersistenceManager getPersistenceManager() {
        return getPersistenceManager(null);
    }

    /** Get an instance of PersistenceManager from this factory.  The instance has
     * default values for options.  The parameters userid and password are used
     * when obtaining datastore connections from the connection pool.
     * Invokes <code>init</code> at the first call.
     *
     * @param userid the userid for the connection
     * @param password the password for the connection
     * @return a PersistenceManager instance with default options.
     */
    public PersistenceManager getPersistenceManager(String userid, String password) {
        Object cs = null;
        if (csf == null) {
            String cn = connectionProperties.getProperty(Speedo.CONNECTION_SPEC_FACTORY);
            if (cn != null && cn.length() > 0) {
                try {
                    csf = (ConnectionSpecFactory) Class.forName(cn).newInstance();
                    cs = csf.getConnectionSpec(this, userid, password);
                } catch (Exception e) {
                    logger.log(BasicLevel.ERROR,
                        "Impossible to instanciate the ConnectionSpecFactory: "
                        + cn, e);
                }
            }
        } else {
            cs = csf.getConnectionSpec(this, userid, password);
        }
        return getPersistenceManager(cs);
    }

    public PersistenceManager getPersistenceManager(Object cs) {
        if (!(_this instanceof PersistenceManagerFactory)) {
            SubComponentController parentCtrl = (SubComponentController)
                    ((ComponentIdentity) _this).getFcInterface("subcomponent-controller");
            _this = parentCtrl.getFcParents()[0]
                    .getFcInterface("persistence-manager-factory");
        }
        ProxyManager pm = pms.lookup((PersistenceManagerFactory) _this);
        if (pm != null)
            return pm;
        try {
            // Gets a PM from the pool
            PoolResource pr = managedPM.getResource(null);
            ComponentIdentity ci = ((InterfaceReference) pr).getFcItfOwner();
            pm = (ProxyManager) ci.getFcInterface("persistence-manager");
            pm.open(cs);
            pms.bind(pm);
            return pm;
        } catch (Exception e) {
            throw new JDODataStoreException(
                    "Cannot provides a PersistenceManager ", new Exception[]{e});
        }
    }

    /** Set the user name for the data store connection.
     * @param userName the user name for the data store connection.
     */
    public void setConnectionUserName(String userName) {
        connectionProperties.setProperty("UserName", userName);
    }

    /** Get the user name for the data store connection.
     * @return the user name for the data store connection.
     */
    public String getConnectionUserName() {
        return connectionProperties.getProperty("UserName");
    }

    /** Set the password for the data store connection.
     * @param password the password for the data store connection.
     */
    public void setConnectionPassword(String password) {
        connectionProperties.setProperty("Password", password);
    }

    /** Get the password for the data store connection.
     * @return  password the password for the data store connection.
     */
    protected String getConnectionPassword() {
        return connectionProperties.getProperty("Password");
    }

    /** Set the URL for the data store connection.
     * @param URL the URL for the data store connection.
     */
    public void setConnectionURL(String URL) {
        connectionProperties.setProperty("URL", URL);
    }

    /** Get the URL for the data store connection.
     * @return the URL for the data store connection.
     */
    public String getConnectionURL() {
        return connectionProperties.getProperty("URL");
    }

    /** Set the driver name for the data store connection.
     * @param driverName the driver name for the data store connection.
     */
    public void setConnectionDriverName(String driverName) {
        connectionProperties.setProperty("DriverName", driverName);
    }

    /** Get the driver name for the data store connection.
     * @return the driver name for the data store connection.
     */
    public String getConnectionDriverName() {
        return connectionProperties.getProperty("DriverName");
    }

    /** Set the name for the data store connection factory.
     * @param connectionFactoryName the name of the data store connection factory.
     */
    public void setConnectionFactoryName(String connectionFactoryName) {
        connectionProperties.setProperty(
                "ConnectionFactoryName", connectionFactoryName);
    }

    /** Get the name for the data store connection factory.
     * @return the name of the data store connection factory.
     */
    public String getConnectionFactoryName() {
        return connectionProperties.getProperty("ConnectionFactoryName");
    }

    /** Set the data store connection factory.  JDO implementations
     * will support specific connection factories.  The connection
     * factory interfaces are not part of the JDO specification.
     * @param connectionFactory the data store connection factory.
     */
    public void setConnectionFactory(Object connectionFactory) {
        throw new JDOUnsupportedOptionException("Cannot set a Connection Factory, use <> instead");
    }

    /** Get the data store connection factory.
     * @return the data store connection factory.
     */
    public Object getConnectionFactory() {
        return mapper.getConnectionFactory();
    }

    /** Set the name for the second data store connection factory.  This is
     * needed for managed environments to get nontransactional connections for
     * optimistic transactions.
     * @param connectionFactoryName the name of the data store connection factory.
     */
    public void setConnectionFactory2Name(String connectionFactoryName) {
        connectionProperties.setProperty("ConnectionFactory2Name", connectionFactoryName);
    }

    /** Get the name for the second data store connection factory.  This is
     * needed for managed environments to get nontransactional connections for
     * optimistic transactions.
     * @return the name of the data store connection factory.
     */
    public String getConnectionFactory2Name() {
        return connectionProperties.getProperty("ConnectionFactory2Name");
    }

    /** Set the second data store connection factory.  This is
     * needed for managed environments to get nontransactional connections for
     * optimistic transactions.  JDO implementations
     * will support specific connection factories.  The connection
     * factory interfaces are not part of the JDO specification.
     * @param connectionfactory the data store connection factory.
     */
    public void setConnectionFactory2(Object connectionfactory) {
        throw new JDOUnsupportedOptionException(
                "Cannot set a second Connection Factory");
    }

    /** Get the second data store connection factory.  This is
     * needed for managed environments to get nontransactional connections for
     * optimistic transactions.
     * @return the data store connection factory.
     */
    public Object getConnectionFactory2() {
        throw new JDOUnsupportedOptionException(
                "Cannot get a second Connection Factory");
    }

    /** Set the default Multithreaded setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @param flag the default Multithreaded setting.
     */
    public void setMultithreaded(boolean flag) {
        connectionProperties.setProperty(
                "Multithreaded", new Boolean(flag).toString());
    }

    /** Get the default Multithreaded setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @return the default Multithreaded setting.
     */
    public boolean getMultithreaded() {
        return connectionProperties.getProperty("Multithreaded").equals("true");
    }

    /** Set the default Optimistic setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @param flag the default Optimistic setting.
     */
    public void setOptimistic(boolean flag) {
        connectionProperties.setProperty(
                "Optimistic", new Boolean(flag).toString());
    }

    /** Get the default Optimistic setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @return the default Optimistic setting.
     */
    public boolean getOptimistic() {
        return connectionProperties.getProperty("Optimistic","").equals("true");
    }

    /** Set the default RetainValues setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @param flag the default RetainValues setting.
     */
    public void setRetainValues(boolean flag) {
        connectionProperties.setProperty(
                "RetainValues", new Boolean(flag).toString());
    }

    /** Get the default RetainValues setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @return the default RetainValues setting.
     */
    public boolean getRetainValues() {
        return connectionProperties.getProperty("RetainValues").equals("true");
    }

    /** Set the default NontransactionalRead setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @param flag the default NontransactionalRead setting.
     */
    public void setNontransactionalRead(boolean flag) {
        connectionProperties.setProperty(
                "NonTransactionalRead", new Boolean(flag).toString());
    }

    /** Get the default NontransactionalRead setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @return the default NontransactionalRead setting.
     */
    public boolean getNontransactionalRead() {
        return connectionProperties.getProperty("NonTransactionalRead").equals("true");
    }

    /** Set the default NontransactionalWrite setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @param flag the default NontransactionalWrite setting.
     */
    public void setNontransactionalWrite(boolean flag) {
        connectionProperties.setProperty(
                "NonTransactionalWrite", new Boolean(flag).toString());
    }

    /** Get the default NontransactionalWrite setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @return the default NontransactionalWrite setting.
     */
    public boolean getNontransactionalWrite() {
        return connectionProperties.getProperty("NonTransactionalWrite").equals("true");
    }

    /** Set the default IgnoreCache setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @param flag the default IgnoreCache setting.
     */
    public void setIgnoreCache(boolean flag) {
        connectionProperties.setProperty(
                "IgnoreCache", new Boolean(flag).toString());
    }

    /** Get the default IgnoreCache setting for all PersistenceManager instances
     * obtained from this factory.
     *
     * @return the default IngoreCache setting.
     */
    public boolean getIgnoreCache() {
        return connectionProperties.getProperty("IgnoreCache").equals("true");
    }

    /** Get the MaxPool setting for the PersistenceManager
     * pool for this factory.
     * @return the MaxPool setting.
     */
    public int getMaxPool() {
        return Integer.parseInt(connectionProperties.getProperty("MaxPool"));
    }

    /** Set the MaxPool setting for the PersistenceManager
     * pool for this factory.
     * @param maxPool the MaxPool setting.
     */
    public void setMaxPool(int maxPool) {
        connectionProperties.setProperty("MaxPool", Integer.toString(maxPool));
    }

    /** Get the MinPool setting for the PersistenceManager
     * pool for this factory.
     * @return the MinPool setting.
     */
    public int getMinPool() {
        return Integer.parseInt(connectionProperties.getProperty("MinPool"));
    }

    /** Set the MinPool setting for the PersistenceManager
     * pool for this factory.
     * @param minPool the MinPool setting.
     */
    public void setMinPool(int minPool) {
        connectionProperties.setProperty("MinPool", Integer.toString(minPool));
    }

    /** Get the MsWait setting for the PersistenceManager
     * pool for this factory.
     * @return the MsWait setting.
     */
    public int getMsWait() {
        return Integer.parseInt(connectionProperties.getProperty("MsWait"));
    }

    /** Set the MsWait setting for the PersistenceManager
     * pool for this factory.
     * @param msWait the MsWait setting.
     */
    public void setMsWait(int msWait) {
        connectionProperties.setProperty("MsWait", Integer.toString(msWait));
    }

    /** Return non-configurable properties of this PersistenceManagerFactory.
     * Properties with keys VendorName and VersionNumber are required.  Other
     * keys are optional.
     * @return the non-configurable properties of this
     * PersistenceManagerFactory.
     */
    public Properties getProperties() {
        return connectionProperties;
    }

    /** The application can determine from the results of this
     * method which optional features, and which query languages
     * are supported by the JDO implementation.
     * <P>Each supported JDO optional feature is represented by a
     * String with one of the following values:
     *
     * <P>javax.jdo.option.TransientTransactional
     * <P>javax.jdo.option.NontransactionalRead
     * <P>javax.jdo.option.NontransactionalWrite
     * <P>javax.jdo.option.RetainValues
     * <P>javax.jdo.option.Optimistic
     * <P>javax.jdo.option.ApplicationIdentity
     * <P>javax.jdo.option.DatastoreIdentity
     * <P>javax.jdo.option.NonDatastoreIdentity
     * <P>javax.jdo.option.ArrayList
     * <P>javax.jdo.option.HashMap
     * <P>javax.jdo.option.Hashtable
     * <P>javax.jdo.option.LinkedList
     * <P>javax.jdo.option.TreeMap
     * <P>javax.jdo.option.TreeSet
     * <P>javax.jdo.option.Vector
     * <P>javax.jdo.option.Map
     * <P>javax.jdo.option.List
     * <P>javax.jdo.option.Array
     * <P>javax.jdo.option.NullCollection
     *
     *<P>The standard JDO query language is represented by a String:
     *<P>javax.jdo.query.JDOQL
     * @return the List of String representing the supported Options
     */
    public Collection supportedOptions() {
        return null;
    }

    public void setRestoreValues(boolean b) {
    }

    public boolean getRestoreValues() {
        return false;
    }

    // IMPLEMENTATION OF THE PoolMathcFactory INTERFACE //
    //--------------------------------------------------//

    /**
     * Creates a new <code>PoolResource</code>.
     * This methos is invoked by the owned <code>Pool</code>.
     *
     * @param o params used to build a new PoolResource
     * @return the build <code>SpeedoPersistenceManager</code>
     */
    public PoolResource createResource(Object o) throws Exception {
        logger.log(BasicLevel.DEBUG, "create a new Proxy manager");
        // instantiates a new persistence manager component
        ComponentIdentity pmComp = Fractal.getTemplate(pmTemplate).instantiateFc();

        // adds it to the root speedo component, which is the parent of the
        // persistence manager template component
        SubComponentController pmTemplateScc =
                (SubComponentController) pmTemplate.getFcInterface("subcomponent-controller");
        ComponentIdentity speedoComp = pmTemplateScc.getFcParents()[0];
        //Fractal.getLifeCycleController(speedoComp).stopFc();
        Fractal.getContentController(speedoComp).addFcSubComponent(pmComp);

        // binds the newly created persistence manager component to the persistence
        // manager factory component, by copying the bindings of the persistence
        // manager template component
        Fractal.getBindingController(pmComp).bindFc(
                "persistence-manager-factory",
                Fractal.getBindingController(pmTemplate).lookupFc("persistence-manager-factory"));
        Fractal.getBindingController(pmComp).bindFc(
                "persistence-manager-pool",
                Fractal.getBindingController(pmTemplate).lookupFc("persistence-manager-pool"));


        Fractal.getLifeCycleController(pmComp).startFc();
        //Fractal.getLifeCycleController(speedoComp).startFc();
        return (PoolResource) pmComp.getFcInterface("pool-resource");
    }

    /**
     * Tests wether a <code>PoolResource</code> can be managed by the current Pool.
     * This methos is invoked by the owned <code>Pool</code>.
     *
     * @param resource the <code>PoolResource</code> to be tested
     * @param o o params used to build a new PoolResource
     * @return true if the resource is an instance of <code>PersistenceManagerImml</code>
     */
    public boolean matchResource(PoolResource resource, Object o) {
        return true;
    }


    // IMPLEMENTATION OF THE PoolProvider INTERFACE //
    //----------------------------------------------//

    /**
     * @return the owned <code>Pool</code>
     */
    public Pool getPool() {
        return managedPM;
    }
}
