package org.objectweb.speedo.runtime;

import org.objectweb.speedo.api.ProxyManagerSwitch;
import org.objectweb.speedo.api.ProxyManager;
import org.objectweb.speedo.api.SpeedoException;
import org.objectweb.fractal.api.control.UserBindingController;

import javax.jdo.PersistenceManagerFactory;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.Collections;

/**
 * This class is an implementation of the ProxyManagerSwitch based on the use
 * of a ThreadLocal field. This field contains an instance of ProxyManager of an
 * ArrayList of ProxyManager. Only bind method is synchronized in order to be
 * thread safe. The lookup method is not synchronized in order to increase the
 * efficiency. Indeed it is not necessary to synchronize the read/write
 * accesses. A SpeedoProxyManagerSwitch is a fractal component which have no
 * bindings.
 *
 * @author S.Chassande-Barrioz
 */
public class SpeedoProxyManagerSwitch
        implements ProxyManagerSwitch {

    protected ThreadLocal pms = new ThreadLocal();

    public final static String BIND_ERROR_MSG = "Impossible to associated to ProxyManager linked to the same PersistenceManagerFactory in the same context (thread)";


    // IMPLEMENTATION OF THE ProxyManagerSwitch INTERFACE //
    //----------------------------------------------------------//

    /**
     * @param pmf is persistent manager factory which manages the returned
     * proxy manager.
     * @return the ProxyManager managed by the given persistence manager factory
     * and bound to current the context, or the null value if there is no
     * ProxyManager.
     */
    public ProxyManager lookup(PersistenceManagerFactory pmf) {
        Object o = pms.get();
        if (o == null) {
            return null;
        } else if (o instanceof ProxyManager) {
            if (((ProxyManager) o).getPersistenceManagerFactory() == pmf)
                return (ProxyManager) o;
        } else {
            for (Iterator it = ((List) o).iterator(); it.hasNext();) {
                ProxyManager pm = (ProxyManager) it.next();
                if (pm.getPersistenceManagerFactory() == pmf)
                    return pm;
            }
        }
        return null;
    }

    /**
     * It assignes a ProxyManager to the current context.
     * @param pm is the ProxyManager
     * @exception SpeedoException is thrown when a proxy manager which has the
     * same PersistentManagerFactory is already bound to the current context.
     */
    public synchronized void bind(ProxyManager pm) throws SpeedoException {
        Object o = pms.get();
        if (o == null) {
            pms.set(pm);
        } else if (o instanceof ProxyManager) {
            // There is one ProxyManager, then check if the old and the new
            // proxy manager are manager by the same PMF.
            ProxyManager pm1 = (ProxyManager) o;
            if (pm1.getPersistenceManagerFactory()
                    == pm.getPersistenceManagerFactory())
                throw new SpeedoException(BIND_ERROR_MSG);

            // Put the old and the new in a List
            ArrayList al = new ArrayList(3);
            al.add(pm1);
            al.add(pm);
            pms.set(al);
        } else {
            // The element is a list of ProxyManager instances, then check if
            // there is a proxy manager linked to the same PersistenceManagerFactory
            List l = (List) o;
            for (Iterator it = l.iterator(); it.hasNext();) {
                if (((ProxyManager) it.next()).getPersistenceManagerFactory()
                        == pm.getPersistenceManagerFactory())
                    throw new SpeedoException(BIND_ERROR_MSG);
            }
            //No element with the same pmf, then add the proxy manager
            ((List) o).add(pm);
        }
    }

    /**
     * It clears the list of ProxyManager for the current context.
     */
    public void clear() {
        pms.set(null);
    }

    /**
     * It clears a ProxyManager for the current context.
     */
    public boolean remove(ProxyManager pm) {
        Object o = pms.get();
        if (o instanceof ProxyManager) {
            if (o == pm) {
                pms.set(null);
                return true;
            }
        } else {
            return ((Collection) o).remove(pm);
        }
        return false;
    }

    /**
     * @return all ProxyManager instances bound with the current context
     */
    public Collection entries() {
        Object o = pms.get();
        if (o instanceof ProxyManager)
            return Collections.singletonList(o);
        else
            return (Collection) o;
    }
}
