/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.runtime.genclass.collection;

import org.objectweb.jorm.api.PExceptionIO;
import org.objectweb.jorm.api.PIndexedElem;

import org.objectweb.speedo.runtime.genclass.GenClassElement;
import org.objectweb.speedo.runtime.genclass.PIndexedElemIterator;
import org.objectweb.speedo.runtime.genclass.GenClassAccessor;
import org.objectweb.speedo.api.SpeedoProxy;
import org.objectweb.speedo.api.ProxyManager;

import java.util.Collection;
import java.util.Iterator;

/**
 * @author	S.Chassande-Barrioz
 */
public class CollectionAccessor
        extends GenClassAccessor
        implements Collection {

    /**
     * The ArrayList used to store the indexed elements of the genclass.
     * Then this list contains also the deleted elements.
     */
    protected Collection elements = null;

    /**
     * Instanciates and initializes a new collection with an initial size.
     */
    public CollectionAccessor(CollectionImpl jdoProxy, int size) {
        super(jdoProxy);
        this.elements = jdoProxy.createCollection(size);
    }

    public String toString() {
        return "CollectionAccessor: id=" + jdoProxy.jdoGetGenClassId();
    }

    // ------------------------------------------------------------------------
    // IMPLEMENTATION OF THE Collection INTERFACE
    // ------------------------------------------------------------------------

    public boolean add(Object o) {
        // Put it in the "elements" list
        GenClassElement element = (GenClassElement) createPIndexedElem();
        if (o instanceof SpeedoProxy) {
            ProxyManager pm = (ProxyManager) jdoProxy.jdoGetPersistenceManager();
            jdoProxy.jdoMakePersistent(pm, (SpeedoProxy) o);
        }
        element.setElement(o);
        synchronized (this) {
            return elements.add(element);
        }
    }

    public boolean addAll(Collection c) {
        Iterator it = c.iterator();
        boolean changed = false;
        while (it.hasNext()) {
            changed |= add(it.next());
        }
        return changed;
    }

    public void clear() {
        elements.clear();
        supportDelta = false;
    }

    public boolean contains(Object o) {
        tmpelem.setElement(o);
        Iterator it = elements.iterator();
        while (it.hasNext()) {
            PIndexedElem e = (PIndexedElem) it.next();
            if (e.equals(tmpelem) && e.getElemStatus() != PIndexedElem.ELEM_DELETED) {
                return true;
            }
        }
        return false;
    }

    public boolean containsAll(Collection c) {
        Iterator it = c.iterator();
        while (it.hasNext()) {
            if (!contains(it.next())) {
                return false;
            }
        }
        return true;
    }

    public boolean equals(Object o) {
        return (o instanceof Collection) &&
                ((Collection) o).size() == size() &&
                containsAll((Collection) o);
    }

    public boolean isEmpty() {
        if (elements.isEmpty()) {
            return true;
        }
        Iterator it = elements.iterator();
        // The set is empty if it contains only deleted indexed elements
        while (it.hasNext()) {
            if (((PIndexedElem) it.next()).getElemStatus() != PIndexedElem.ELEM_DELETED) {
                return false;
            }
        }
        return true;
    }

    public Iterator iterator() {
        return new PIndexedElemIterator(elements, this,
                jdoProxy.jdoGetPersistenceManager(), logger);
    }

    public boolean remove(Object o) {
        tmpelem.setElement(o);   // tmp type = etype.
        synchronized (this) {
            Iterator it = elements.iterator();
            while (it.hasNext()) {
                CollectionElem e = (CollectionElem) it.next();
                if (e.equals(tmpelem) && e.getElemStatus() != PIndexedElem.ELEM_DELETED) {
                    e.setStatus(PIndexedElem.ELEM_DELETED);
                    return true;
                }
            }
            return false;
        }
    }

    public boolean removeAll(Collection c) {
        Iterator it = c.iterator();
        boolean removed = true;
        while (it.hasNext()) {
            removed &= remove(it.next());
        }
        return removed;
    }

    public boolean retainAll(Collection c) {
        throw new UnsupportedOperationException();
    }

    public int size() {
        Iterator it = elements.iterator();
        int size = 0;
        while (it.hasNext()) {
            if (((PIndexedElem) it.next()).getElemStatus() != PIndexedElem.ELEM_DELETED) {
                size++;
            }
        }
        return size;
    }

    public Object[] toArray() {
        throw new UnsupportedOperationException();
    }

    public Object[] toArray(Object[] a) {
        throw new UnsupportedOperationException();
    }

    // ------------------------------------------------------------------------
    // IMPLEMENTATION OF THE PGenClassAccessor INTERFACE
    // ------------------------------------------------------------------------

    public PIndexedElem createPIndexedElem() {
        return new CollectionElem();
    }

    public void paAdd(PIndexedElem elem, Object conn) throws PExceptionIO {
        // the elem is read from the DS, set it to unmodified.
        ((CollectionElem) elem).setStatus(PIndexedElem.ELEM_UNMODIFIED);
        elements.add(elem);
    }

    public int paGetNbElem() {
        return size();
    }

    public Iterator paIterator() {
        return elements.iterator();
    }

    public void paSetNbElem(int nbelem) {
        elements = ((CollectionImpl) jdoProxy).createCollection(nbelem);
    }
}
