/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
package org.objectweb.speedo.runtime.genclass.collection;

import org.objectweb.speedo.api.SpeedoAccessor;
import org.objectweb.speedo.api.ProxyManager;
import org.objectweb.speedo.api.SpeedoProxy;
import org.objectweb.speedo.api.SpeedoGenClassProxy;
import org.objectweb.speedo.runtime.genclass.GenClass;
import org.objectweb.jorm.type.api.PType;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

/**
 * This class is an implementation of the java.util.Collection interface.
 *
 * @author	Eric.Bruneton, S.Chassande-Barrioz
 */
public class CollectionImpl extends GenClass implements Collection {

    private final static int DEFAULT_SIZE = -1;

    /**
     * is the internal collection used when the generic class is not persistent
     */
    protected Collection collection = null;

    /**
     * Instantiates a new collection.
     */
    public CollectionImpl() {
        this(DEFAULT_SIZE);
    }

    /**
     * Instanciates and initializes a new collection with an initial size.
     */
    public CollectionImpl(int size) {
        super();
        collection = createCollection(size);
    }

    public CollectionImpl(Collection collection,
                          PType ptype,
                          String linkedField,
                          PName linkedPName,
                          String mapperName) {
        super(ptype, linkedField, linkedPName, mapperName);
        this.collection = createCollection(
                (collection == null ? DEFAULT_SIZE : collection.size()));
        setInner(collection);
    }

    public CollectionImpl(PType ptype,
                          String linkedField,
                          PName linkedPName,
                          String mapperName) {
        this(null, ptype, linkedField, linkedPName, mapperName);
    }

    public String toString() {
        return "CollectionImpl: id=" + jdoGetGenClassId();
    }

    // ------------------------------------------------------------------------
    // IMPLEMENTATION OF THE Collection INTERFACE
    // ------------------------------------------------------------------------

    public boolean add(Object o) {
        boolean result;
        if (!jdoIsActive) {
            result = collection.add(o);
        } else {
            CollectionAccessor ca = (CollectionAccessor) getSpeedoAccessor();
            jdoWriteIntention(ca);
            result = ca.add(o);
        }
        return result;
    }

    public boolean addAll(Collection c) {
        boolean result;
        if (!jdoIsActive) {
            result = collection.addAll(c);
        } else {
            CollectionAccessor ca = (CollectionAccessor) getSpeedoAccessor();
            jdoWriteIntention(ca);
            result = ca.addAll(c);
        }
        return result;
    }

    public void clear() {
        if (!jdoIsActive) {
            collection.clear();
        } else {
            CollectionAccessor ca = (CollectionAccessor) getSpeedoAccessor();
            jdoWriteIntention(ca);
            ca.clear();
        }
    }

    public boolean contains(Object o) {
        if (!jdoIsActive) {
            return collection.contains(o);
        } else {
            CollectionAccessor ca = (CollectionAccessor) getSpeedoAccessor();
            jdoReadIntention(ca);
            return ca.contains(o);
        }
    }

    public boolean containsAll(Collection c) {
        if (!jdoIsActive) {
            return collection.containsAll(c);
        } else {
            CollectionAccessor ca = (CollectionAccessor) getSpeedoAccessor();
            jdoReadIntention(ca);
            return ca.containsAll(c);
        }
    }

    public boolean equals(Object o) {
        if (!jdoIsActive) {
            return collection.equals(o);
        } else {
            CollectionAccessor ca = (CollectionAccessor) getSpeedoAccessor();
            jdoReadIntention(ca);
            return ca.equals(o);
        }
    }

    public boolean isEmpty() {
        if (!jdoIsActive) {
            return collection.isEmpty();
        } else {
            CollectionAccessor ca = (CollectionAccessor) getSpeedoAccessor();
            jdoReadIntention(ca);
            return ca.isEmpty();
        }
    }

    public Iterator iterator() {
        if (!jdoIsActive) {
            return collection.iterator();
        } else {
            CollectionAccessor ca = (CollectionAccessor) getSpeedoAccessor();
            jdoReadIntention(ca);
            return ca.iterator();
        }
    }

    public boolean remove(Object o) {
        boolean result;
        if (!jdoIsActive) {
            result = collection.remove(o);
        } else {
            CollectionAccessor ca = (CollectionAccessor) getSpeedoAccessor();
            jdoWriteIntention(ca);
            result = ca.remove(o);
        }
        return result;
    }

    public boolean removeAll(Collection c) {
        boolean result;
        if (!jdoIsActive) {
            result = collection.removeAll(c);
        } else {
            CollectionAccessor ca = (CollectionAccessor) getSpeedoAccessor();
            jdoWriteIntention(ca);
            result = ca.removeAll(c);
        }
        return result;
    }

    public boolean retainAll(Collection c) {
        throw new UnsupportedOperationException();
    }

    public int size() {
        if (!jdoIsActive) {
            return collection.size();
        } else {
            CollectionAccessor ca = (CollectionAccessor) getSpeedoAccessor();
            jdoReadIntention(ca);
            return ca.size();
        }
    }

    public Object[] toArray() {
        throw new UnsupportedOperationException();
    }

    public Object[] toArray(Object[] a) {
        throw new UnsupportedOperationException();
    }

    // IMPLEMENTATION OF THE SpeedoProxy INTERFACE //
    //---------------------------------------------//

    public void jdoLoadFieldsInAccessor(SpeedoAccessor sa) {
        CollectionAccessor ca = (CollectionAccessor) sa;
        ca.clear();
        ca.addAll(collection);
    }

    public void jdoLoadFieldsFromAccessor(SpeedoAccessor sa) {
        collection.clear();
        collection.addAll((CollectionAccessor) sa);
    }

    public void jdoMakePersistent(ProxyManager pm) {
        if (collection.size() == 0) {
            return;
        }
        Iterator i = collection.iterator();
        while (i.hasNext()) {
            Object elem = i.next();
            if (elem instanceof SpeedoProxy) {
                jdoMakePersistent(pm, (SpeedoProxy) elem);
            } else {
                // Optimisation: If the first element is not a SpeedoProxy, that
                // means that ALL collection elements are not SpeedoProxy
                // implementation
                return;
            }
        }
    }

    public void jdoMakePersistent(ProxyManager pm, SpeedoProxy sp) {
        if (sp instanceof SpeedoGenClassProxy) {
            SpeedoGenClassProxy sgcp = (SpeedoGenClassProxy) sp;
            sgcp.jdoSetMapperName(jdoGetMapperName());
            if (sgcp.jdoGetPType() == null) {
                //Assign the PType to the generic class
                sgcp.jdoSetPType(jdoGetPType().getNestedPType());
            }
            sgcp.jdoSetLinkedField(jdoGetGenClassId());
            sgcp.jdoSetPNameHints(getPName());
        }
        pm.makePersistent(sp);
    }

    // OTHER METHODS //
    // --------------//

    protected SpeedoAccessor createAccessor() {
        return new CollectionAccessor(this, collection.size());
    }

    protected Collection createCollection(int size) {
        return (size == DEFAULT_SIZE ?  new ArrayList() : new ArrayList(size));
    }

    protected CollectionImpl createInstance(int size) {
        return new CollectionImpl(size);
    }

    public Collection getGenClassElements(SpeedoAccessor sa) {
        return ((CollectionAccessor) sa).elements;
    }

    public void setGCValue(Object o) {
        if (jdoIsActive) {
            CollectionAccessor ca = (CollectionAccessor) getSpeedoAccessor();
            ca.clear();
            if (o != null) {
                ca.addAll(((Collection) o));
            }
        } else {
            collection.clear();
            if (o != null) {
                collection.addAll(((Collection) o));
            }
        }
    }


}
