/**
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package org.objectweb.speedo.runtime.persistence;

import org.objectweb.jorm.api.PException;
import org.objectweb.jorm.api.PMapper;
import org.objectweb.jorm.naming.api.PName;
import org.objectweb.perseus.cache.api.CacheEntry;
import org.objectweb.perseus.cache.api.CacheException;
import org.objectweb.perseus.concurrency.api.ConcurrencyManager;
import org.objectweb.perseus.persistence.api.PersistenceException;
import org.objectweb.perseus.persistence.api.TransactionalPersistenceManager;
import org.objectweb.perseus.persistence.api.WorkingSet;
import org.objectweb.perseus.persistence.lib.AbstractPersistenceManager;
import org.objectweb.speedo.api.LifeCycle;
import org.objectweb.speedo.api.ProxyManager;
import org.objectweb.speedo.api.SpeedoAccessor;
import org.objectweb.speedo.api.SpeedoProxy;
import org.objectweb.speedo.api.Transaction;
import org.objectweb.speedo.api.JormFactory;
import org.objectweb.speedo.api.WorkingSetLifeCycle;
import org.objectweb.speedo.api.SpeedoGenClassProxy;
import org.objectweb.speedo.api.ExceptionHelper;
import org.objectweb.speedo.runtime.SpeedoTransaction;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.medor.eval.prefetch.api.PrefetchCache;

import javax.jdo.JDOException;
import javax.resource.ResourceException;
import java.util.Iterator;
import java.sql.SQLException;

public class SpeedoTransactionalPersistenceManager
        extends AbstractPersistenceManager implements TransactionalPersistenceManager {

    public final static String MAPPER_BINDING = "mapper";
    public final static String JORM_FACTORY_BINDING = "jorm-factory";

    protected PMapper mapper = null;
    protected JormFactory jf = null;


    // IMPLEMENTATION OF THE UserBindingController INTERFACE //
    //-------------------------------------------------------//

    public Object getFcBindings(String s) {
        if (MAPPER_BINDING.equals(s))
            return mapper;
        else if (JORM_FACTORY_BINDING.equals(s))
            return jf;
        else
            return super.getFcBindings(s);
    }

    public void addFcBinding(String s, Object o) {
        if (MAPPER_BINDING.equals(s))
            mapper = (PMapper) o;
        else if (JORM_FACTORY_BINDING.equals(s))
            jf = (JormFactory) o;
        else
            super.addFcBinding(s, o);
    }

    public void removeFcBinding(String s, Object o) {
        if (MAPPER_BINDING.equals(s))
            mapper = null;
        else if (JORM_FACTORY_BINDING.equals(s))
            jf = null;
        else
            super.removeFcBinding(s, o);

    }

    // IMPLEMENTATION OF THE AbstractPersistenceManager CLASS //
    //--------------------------------------------------------//

    protected CacheEntry bind(CacheEntry cacheEntry, WorkingSet workingSet)
            throws PersistenceException {
        SpeedoProxy sp = null;
        SpeedoAccessor sa = null;
        if (cacheEntry instanceof SpeedoProxy) {
            sp = (SpeedoProxy) cacheEntry;
            sa = sp.getSpeedoAccessor();
        } else if (cacheEntry instanceof SpeedoAccessor) {
            sa = (SpeedoAccessor) cacheEntry;
            sp = sa.jdoGetSpeedoProxy();
        } else {
            logger.log(BasicLevel.DEBUG, "unmanaged entry " + cacheEntry);
            throw new PersistenceException(
                    "Unmanaged CacheEntry implementation: " + cacheEntry);
        }
        if (workingSet.lookup(sa) == null) {
            Transaction tx = (Transaction) workingSet;
            if (tx.isActive() && tx.getOptimistic()) {
                logger.log(BasicLevel.DEBUG, "Load the value from the proxy into the accessor");
                sp.jdoLoadFieldsInAccessor(sa);
            }
            logger.log(BasicLevel.DEBUG, "bind an entry " + sa +" into a context: " + tx.entries());
            logger.log(BasicLevel.DEBUG, "contains: " + tx.entries().contains(sa));
            workingSet.bind(sa);
            try {
                cache.fix(cacheEntry);
            } catch (CacheException e) {
                throw new PersistenceException(e);
            }
            sa.jdoSetTransaction((Transaction) workingSet);
        }
        return sa;
    }

    protected void initMemoryInstance(Object o, Object oid)
            throws PersistenceException {
        SpeedoProxy sp = (SpeedoProxy) o;
        try {
            sp.init(jf.getPClassMapping(sp));
            sp.bind((PName) oid);
        } catch (PException e) {
            throw new PersistenceException(
                    "Impossible to initialize a new instance", e);
        }
    }

    protected Object getDataAccess(Object context) throws PersistenceException {
        Transaction tx = (Transaction) context;
        Object da = tx.getDataAccess();
        if (da == null) {
            Object cs = ((ProxyManager) tx.getPersistenceManager()).getConnectionSpec();
            try {
                if (cs == null) {
                    da = mapper.getConnection();
                } else {
                    da = mapper.getConnection(cs);
                }
            } catch (PException e) {
                throw new PersistenceException("Impossible to fetch a data access", e);
            }
            ((Transaction) context).assignDataAccess(da);
        }
        return da;
    }

    protected void closeDataAccess(Transaction tx, Object da) throws PersistenceException {
        if (da == null) {
            return;
        }
        tx.removeDataAccess(da);
        try {
            mapper.closeConnection(da);
        } catch (PException e) {
            throw new PersistenceException("Impossible to fetch a data access", e);
        }
    }

    protected void releaseDataAccess(Object context, Object da) throws PersistenceException {
    }

    protected ConcurrencyManager getConcurrencyManager(WorkingSet workingSet) {
        return ((Transaction) workingSet).getOptimistic() ? ocm : cm;
    }

    // IMPLEMENTATION OF THE TransactionalPersistenceManager INTERFACE //
    //-----------------------------------------------------------------//

    public Object createContext() throws PersistenceException {
        return new SpeedoTransaction();
    }

    public void begin(Object context) throws PersistenceException {
        Transaction tx = (Transaction) context;
        byte newStatus = WorkingSetLifeCycle.getNextStatus(
                tx.getStatus(), WorkingSetLifeCycle.BEGIN_TX_ACTION);
        // 1st step: flush objects in working set, outside transaction
        Object da = getDataAccess(tx);
        tx.assignDataAccess(da);
        if (!tx.isManagedEnv() && da != null) {
            //Open a local transaction
            if (da instanceof java.sql.Connection) {
                try {
                    ((java.sql.Connection) da).setAutoCommit(false);
                } catch (SQLException e) {
                    throw new JDOException("", ExceptionHelper.getNested(e));
                }
            } else if (da instanceof javax.resource.cci.Connection) {
                try {
                    ((javax.resource.cci.Connection) da)
                        .getLocalTransaction().begin();
                } catch (ResourceException e) {
                    throw new JDOException("", ExceptionHelper.getNested(e));
                }
            }
        }
        dom.flush(da, tx.entries());
        // 2nd step: change transaction's status
        tx.setStatus(newStatus);
    }

    public void commit(Object context) throws PersistenceException {
        Transaction tx = (Transaction) context;
        PrefetchCache pc = mapper.getPrefetchCache();
        if (pc != null) {
            pc.invalidatePrefetchBuffer(tx);
        }
        ConcurrencyManager cm = getConcurrencyManager(tx);
        byte newStatus = WorkingSetLifeCycle.getNextStatus(
                tx.getStatus(), WorkingSetLifeCycle.COMMIT_TX_ACTION);
        // try to validate transaction with concurrency manager
        boolean valid = cm.validate(tx);
        if (valid) {
            // if ok, notify concurrency manager,
            cm.finalize(tx);
            // flush dirty objects and write them to data support,
            Object da = tx.getDataAccess();
            try {
                dom.flush(da, tx.entries());
                if (!tx.isManagedEnv() && da != null) {
                    //Open a local transaction
                    if (da instanceof java.sql.Connection) {
                        try {
                            ((java.sql.Connection) da).commit();
                            ((java.sql.Connection) da).setAutoCommit(true);
                        } catch (SQLException e) {
                            throw new JDOException("", ExceptionHelper.getNested(e));
                        }
                    } else if (da instanceof javax.resource.cci.Connection) {
                        try {
                            ((javax.resource.cci.Connection) da)
                                .getLocalTransaction().commit();
                        } catch (ResourceException re) {
                            throw new JDOException("", ExceptionHelper.getNested(re));
                        }
                    }
                }
            } finally {
                closeDataAccess(tx, da);
            }
            // copies accessor states into proxies,
            if (tx.getOptimistic()) {
                for (Iterator i = tx.entries().iterator(); i.hasNext();) {
                    SpeedoAccessor sa = (SpeedoAccessor) i.next();
                    SpeedoProxy sp = sa.jdoGetSpeedoProxy();
                    if (LifeCycle.isDirty(sa.jdoGetStatus())) {
                        sp.jdoLoadFieldsFromAccessor(sa);
                    }
                    sp.removeSpeedoAccessor(sa);
                }
            }
            // and, finally, change transaction's status
            tx.setStatus(newStatus);
            unfixTxElements(tx);
            tx.clear();
        } else {
            rollback(tx);
        }
    }

    public void rollback(Object context) throws PersistenceException {
        Transaction tx = (Transaction) context;
        PrefetchCache pc = mapper.getPrefetchCache();
        if (pc != null) {
            pc.invalidatePrefetchBuffer(tx);
        }
        byte newStatus = WorkingSetLifeCycle.getNextStatus(
                tx.getStatus(), WorkingSetLifeCycle.ROLLBACK_TX_ACTION);
        // notify concurrency manager
        getConcurrencyManager(tx).abort(tx);
        // reload the state of dirty objects from data support
        if (!tx.getOptimistic()) {
            Object da = tx.getDataAccess();
            try {
                for (Iterator i = tx.entries().iterator(); i.hasNext();) {
                    SpeedoAccessor sa = (SpeedoAccessor) i.next();
                    if (LifeCycle.isDirty(sa.jdoGetStatus())) {
                        sm.read(da, sa.getCeIdentifier(), sa.getCeObject(), tx);
                    }
                }
            } finally {
                if (!tx.isManagedEnv() && da != null) {
                    //Open a local transaction
                    if (da instanceof java.sql.Connection) {
                        try {
                            ((java.sql.Connection) da).rollback();
                            ((java.sql.Connection) da).setAutoCommit(true);
                        } catch (SQLException e) {
                            throw new JDOException("", ExceptionHelper.getNested(e));
                        }
                    } else if (da instanceof javax.resource.cci.Connection) {
                        try {
                            ((javax.resource.cci.Connection) da)
                                .getLocalTransaction().rollback();
                        } catch (ResourceException re) {
                            throw new JDOException("", ExceptionHelper.getNested(re));
                        }
                    }
                }
                closeDataAccess(tx, da);
            }
        } else {
            for (Iterator i = tx.entries().iterator(); i.hasNext();) {
                SpeedoAccessor sa = (SpeedoAccessor) i.next();
                SpeedoProxy sp = sa.jdoGetSpeedoProxy();
                sp.removeSpeedoAccessor(sa);
            }
        }
        // flush dirty objects WITHOUT writing them to data support
        dom.makeClean(tx.entries());
        // change transaction's status
        tx.setStatus(newStatus);
        unfixTxElements(tx);
        tx.clear();
    }

    public void close(Object context) throws PersistenceException {
        Transaction tx = (Transaction) context;
        PrefetchCache pc = mapper.getPrefetchCache();
        if (pc != null) {
            pc.invalidatePrefetchBuffer(tx);
        }
        byte newStatus = WorkingSetLifeCycle.getNextStatus(
                tx.getStatus(), WorkingSetLifeCycle.CLOSE_ACTION);

        if (!tx.isActive()) {
            boolean valid = cm.validate(tx);
            if (valid) {
                // if ok, notify concurrency manager,
                cm.finalize(tx);
                // flush dirty objects and write them to data support,
                Object da = tx.getDataAccess();
                try {
                    dom.flush(da, tx.entries());
                } finally {
                    closeDataAccess(tx, da);
                }
            } else {
                // flush dirty objects WITHOUT writing them to data support
                for (Iterator i = tx.entries().iterator(); i.hasNext();) {
                    SpeedoAccessor sa = (SpeedoAccessor) i.next();
                    SpeedoProxy sp = sa.jdoGetSpeedoProxy();
                    sp.removeSpeedoAccessor(sa);
                }
            }
            dom.makeClean(tx.entries());
            unfixTxElements(tx);
            tx.clear();
        }
        tx.setStatus(newStatus);
    }

    public byte getStatus(Object context) {
        return ((Transaction) context).getStatus();
    }

    private void unfixTxElements(Transaction tx) throws PersistenceException {
        try {
            Iterator it = tx.entries().iterator();
            while(it.hasNext()) {
                SpeedoProxy sp = ((SpeedoAccessor) it.next()).jdoGetSpeedoProxy();
                if (!(sp instanceof SpeedoGenClassProxy)) {
                    logger.log(BasicLevel.DEBUG, "unfix: " + sp + " " + sp.getClass().getName());
                    cache.unfix(sp);
                }
            }
        } catch (CacheException e) {
            throw new PersistenceException(e);
        }
    }
}
