/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 0.1
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: P. Dechamboux, Y. Devaux, F. Dhumeaux.
 *
 */

package org.objectweb.speedo.runtime.query;

import org.objectweb.speedo.runtime.query.api.QueryCompiler;
import org.objectweb.speedo.runtime.query.api.QueryDefinition;
import org.objectweb.speedo.runtime.query.api.QueryManager;
import org.objectweb.speedo.api.ProxyManager;
import org.objectweb.util.monolog.api.BasicLevel;
import org.objectweb.util.monolog.api.Logger;
import org.objectweb.util.monolog.api.LoggerFactory;
import org.objectweb.perseus.pool.api.PoolResource;
import org.objectweb.perseus.pool.api.Pool;

import javax.jdo.Extent;
import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOUserException;
import javax.jdo.PersistenceManager;
import javax.jdo.Query;
import javax.jdo.JDOException;
import java.util.Collection;
import java.util.Map;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.List;
import java.util.ArrayList;

/**
 * SpeedoQuery is the basic implementation of the javax.jdo.Query interface.
 * This class is just a wrapper to a SpeedoQueryCompiler object which is a
 * reused object.
 *
 * @author	F. Dhumeaux, X. Spengler
 */
public class SpeedoQuery
        implements Query, QueryDefinition, PoolResource {

    public final static String LOGGER_NAME
            = "org.objectweb.speedo.rt.query.SpeedoQuery";

    /**
     * IgnoreCache option.
     * The ignoreCache option setting specifies whether the query should execute
     * entirely in the back end, instead of in the cache.  If this flag is set
     * to true, an implementation might be able to optimize the query
     * execution by ignoring changed values in the cache.  For optimistic
     * transactions, this can dramatically improve query response times.
     */
    private boolean ignoreCache = true;

    /**
     * Candidate classes or extent for the query.
     */
    private Collection candidateInstances = null;
    private Extent extentClass = null;

    /**
     * Class of candadate classes.
     */
    private Class candidateClass = null;

    /**
     * Query filter.
     */
    private String filter = "true";

    /**
     * Query parameters and variables declaration
     */
    private String parameters = null;
    private String variables = null;

    /**
     * the import statements is transformed into a simple vector
     */
    private List importStatements = null;
    private List order = null;

    /**
     * The queryManager object is used to create or return a QueryCompiler
     * object.
     */
    private QueryManager queryManager = null;

    /**
     * Logger for monolog.
     */
    private Logger logger = null;

    private boolean hasChanged = true;

    /**
     * PersistenceManager which manages this query.
     */
    private ProxyManager pm;

    private List result;

    /**
     * sQueryCompiler is the query, this object can compile and execute the
     * query. This object is reused in the past, or future.
     */
    private QueryCompiler qc = null;

    private LoggerFactory loggerFactory = null;

    private Pool pool = null;

    public SpeedoQuery() {
        result = new ArrayList(1);
    }

    public QueryManager getQueryManager() {
        return queryManager;
    }

    public void setQueryManager(QueryManager _qm) {
        queryManager = _qm;
    }

    public QueryCompiler getQueryCompiler() {
        return qc;
    }

    public void setQueryCompiler(QueryCompiler _qc) {
        qc = _qc;
    }

    public void setProxyManager(ProxyManager _pm) {
        pm = _pm;
    }

    public LoggerFactory getLoggerFactory() {
        return loggerFactory;
    }

    public void setLoggerFactory(LoggerFactory lf) {
        loggerFactory = lf;
        if (loggerFactory != null) {
            logger = loggerFactory.getLogger(LOGGER_NAME);
        }
    }

    public void defineWith(QueryDefinition sq) {
        if (sq == null) {
            parameters = null;
            variables = null;
            importStatements = null;
            ignoreCache = true;
            order = null;
            filter = "true";
            candidateClass = null;
            hasChanged = true;
        } else {
            // data from the existing query object
            parameters = sq.getParameters();
            variables = sq.getVariables();
            importStatements = sq.getImportStatements();
            ignoreCache = sq.isIgnoreCache();
            order = sq.getOrder();
            filter = sq.getFilter();
            candidateClass = sq.getCandidateClass();
        }
    }

    public void setPool(Pool pool) {
        this.pool = pool;
    }

    // IMPLEMENTATION OF THE PoolResource INTERFACE //
    //----------------------------------------------//

    public void destroy() throws Exception {
        pool = null;
    }

    // IMPLEMENTATION OF THE Query INTERFACE //
    //---------------------------------------//

    /**
     * Set the class of the candidate instances of the query.
     * <P>The class specifies the class
     * of the candidates of the query.  Elements of the candidate collection
     * that are of the specified class are filtered before being
     * put into the result Set.
     * @param cls the Class of the candidate instances.
     */
    public void setClass(Class cls) {
        assertPMIsOpen();
        if (candidateClass != null) hasChanged = true;
        candidateClass = cls;
        assertCandidateClass();
    }

    /**
     * Set the candidate Extent to query.
     * @param pcs the Candidate Extent.
     */
    public void setCandidates(Extent pcs) {
        assertPMIsOpen();
        if (extentClass != null) hasChanged = true;
        extentClass = pcs;
        assertExtentClass();
    }

    /**
     * Set the candidate Set to query.
     * @param pcs the Candidate collection.
     */
    public void setCandidates(Collection pcs) {
        assertPMIsOpen();
        if (candidateInstances != null) hasChanged = true;
        candidateInstances = pcs;
        assertCandidateInstances();
    }

    /**
     * Set the filter for the query.
     * <P>The filter specification is a String containing a boolean
     * expression that is to be evaluated for each of the instances
     * in the candidate collection. If the filter is not specified,
     * then it defaults to "true", which has the effect of filtering
     * the input Set only for class type.
     * <P>An element of the candidate collection is returned in the result if:
     * <ul><li>it is assignment compatible to the candidate Class of the Query; and
     * <li>for all variables there exists a value for which the filter
     * expression evaluates to true.
     * </ul>
     * <P>The user may denote uniqueness in the filter expression by
     * explicitly declaring an expression (for example, e1 != e2).
     * <P>Rules for constructing valid expressions follow the Java
     * language, except for these differences:
     * <ul>
     * <li>Equality and ordering comparisons between primitives and instances
     * of wrapper classes are valid.
     * <li>Equality and ordering comparisons of Date fields and Date
     * parameters are valid.
     * <li>White space (non-printing characters space, tab, carriage
     * return, and line feed) is a separator and is otherwise ignored.
     * <li>The assignment operators =, +=, etc. and pre- and post-increment
     * and -decrement are not supported. Therefore, there are no side
     * effects from evaluation of any expressions.
     * <li>Methods, including object construction, are not supported, except
     * for Set.contains(Object o), Set.isEmpty(),
     * String.startsWith (String s), and String.endsWith (String e).
     * Implementations might choose to support non-mutating method
     * calls as non-standard extensions.
     * <li>Navigation through a null-valued field, which would throw
     * NullPointerException, is treated as if the filter expression
     * returned false for the evaluation of the current set of variable
     * values. Other values for variables might still qualify the candidate
     * instance for inclusion in the result set.
     * <li>Navigation through multi-valued fields (Set types) is
     * specified using a variable declaration and the
     * Set.contains(Object o) method.
     * </ul>
     * <P>Identifiers in the expression are considered to be in the name
     * space of the specified class, with the addition of declared imports,
     * parameters and variables. As in the Java language, this is a reserved
     * word which means the element of the collection being evaluated.
     * <P>Navigation through single-valued fields is specified by the Java
     * language syntax of field_name.field_name....field_name.
     * <P>A JDO implementation is allowed to reorder the filter expression
     * for optimization purposes.
     * @param f the query filter.
     */
    public void setFilter(String f) {
        assertPMIsOpen();
        if (f == null) {
            filter = "true";
        } else {
            filter = f;
            hasChanged = true;
        }
    }

    /**
     * Set the import statements to be used to identify the fully qualified name of
     * variables or parameters.  Parameters and unbound variables might
     * come from a different class from the candidate class, and the names
     * need to be declared in an import statement to eliminate ambiguity.
     * Import statements are specified as a String with semicolon-separated
     * statements.
     * <P>The String parameter to this method follows the syntax of the
     * import statement of the Java language.
     * @param imports import statements separated by semicolons.
     */
    public void declareImports(String imports) {
        assertPMIsOpen();
        if (importStatements != null) hasChanged = true;
        StringTokenizer tok = new StringTokenizer(imports, ";");
        importStatements = new Vector();
        while (tok.hasMoreTokens())
            importStatements.add(tok.nextToken().trim());
    }

    /**
     * Declare the list of parameters query execution.
     * The parameter declaration is a String containing one or more query
     * parameter declarations separated with commas. Each parameter named
     * in the parameter declaration must be bound to a value when
     * the query is executed.
     * <P>The String parameter to this method follows the syntax for formal
     * parameters in the Java language.
     * @param parameters the list of parameters separated by commas.
     */
    public void declareParameters(String parameters) {
        assertPMIsOpen();
        if (this.parameters != null) hasChanged = true;
        this.parameters = parameters;
        assertParameters();
    }

    /**
     * Declare the unbound variables to be used in the query. Variables
     * might be used in the filter, and these variables must be declared
     * with their type. The unbound variable declaration is a String
     * containing one or more unbound variable declarations separated
     * with semicolons. It follows the syntax for local variables in
     * the Java language.
     * @param variables the variables separated by semicolons.
     */
    public void declareVariables(String variables) {
        assertPMIsOpen();
        if (this.variables != null) hasChanged = true;
        this.variables = variables;
        assertVariables();
    }

    /**
     * Set the ordering specification for the result Set.  The
     * ordering specification is a String containing one or more ordering
     * declarations separated by commas.
     *
     * <P>Each ordering declaration is the name of the field on which
     * to order the results followed by one of the following words:
     * "ascending" or "descending".
     *
     *<P>The field must be declared in the candidate class or must be
     * a navigation expression starting with a field in the candidate class.
     *
     *<P>Valid field types are primitive types except boolean; wrapper types
     * except Boolean; BigDecimal; BigInteger; String; and Date.
     * @param ordering the ordering specification.
     */
    public void setOrdering(String ordering) {
        assertPMIsOpen();
        if (order != null) hasChanged = true;
        StringTokenizer tok = new StringTokenizer(ordering, ",");
        order = new ArrayList();
        while (tok.hasMoreTokens())
            order.add(tok.nextToken().trim());
    }

    /**
     * Set the ignoreCache option.  The default value for this option was
     * set by the PersistenceManagerFactory or the PersistenceManager used
     * to create this Query.
     *
     * The ignoreCache option setting specifies whether the query should execute
     * entirely in the back end, instead of in the cache.  If this flag is set
     * to true, an implementation might be able to optimize the query
     * execution by ignoring changed values in the cache.  For optimistic
     * transactions, this can dramatically improve query response times.
     * @param ignoreCache the setting of the ignoreCache option.
     */
    public void setIgnoreCache(boolean ignoreCache) {
        assertPMIsOpen();
        this.ignoreCache = ignoreCache;
    }

    /**
     * Get the ignoreCache option setting.
     * @return the ignoreCache option setting.
     * @see #setIgnoreCache
     */
    public boolean getIgnoreCache() {
        assertPMIsOpen();
        return ignoreCache;
    }

    /**
     * Verify the elements of the query and provide a hint to
     * the query to prepare and optimize an execution plan.
     */
    public void compile() {
        // build a QueryCompiler object with the current object.
        // if the object already exists, it is returned
        qc = queryManager.getQueryCompiler(this);

        // compile the QueryCompiler object
        try {
            qc.compile();
        } catch (Exception e) {
            throw new JDOException("Impossible to compile a query ", e);
        }
        hasChanged = false;
    }

    /**
     * Execute the query and return the filtered Set.
     * @return the filtered Set.
     * @see #executeWithArray(Object[] parameters)
     */
    public Object execute() {
        beforeExecute(0);
        try {
            Object res = qc.execute((Object[]) null, pm);
            result.add(res);
            return res;
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "execute failed", e);
            throw new JDOException("Impossible to execute a query ", e);
        }
    }

    /**
     * Execute the query and return the filtered Set.
     * @return the filtered Set.
     * @see #executeWithArray(Object[] parameters)
     * @param p1 the value of the first parameter declared.
     */
    public Object execute(Object p1) {
        beforeExecute(1);
        try {
            Object res = qc.execute(new Object[]{p1}, pm);
            result.add(res);
            return res;
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "execute failed", e);
            throw new JDOException("Impossible to execute a query ", e);
        }
    }

    /**
     * Execute the query and return the filtered Set.
     * @return the filtered Set.
     * @see #executeWithArray(Object[] parameters)
     * @param p1 the value of the first parameter declared.
     * @param p2 the value of the second parameter declared.
     */
    public Object execute(Object p1, Object p2) {
        beforeExecute(2);
        try {
            Object res = qc.execute(new Object[]{p1, p2}, pm);
            result.add(res);
            return res;
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "execute failed", e);
            throw new JDOException("Impossible to execute a query ", e);
        }
    }

    /**
     * Execute the query and return the filtered Set.
     * @return the filtered Set.
     * @see #executeWithArray(Object[] parameters)
     * @param p1 the value of the first parameter declared.
     * @param p2 the value of the second parameter declared.
     * @param p3 the value of the third parameter declared.
     */
    public Object execute(Object p1, Object p2, Object p3) {
        beforeExecute(3);
        try {
            Object res = qc.execute(new Object[]{p1, p2, p3}, pm);
            result.add(res);
            return res;
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "execute failed", e);
            throw new JDOException("Impossible to execute a query ", e);
        }
    }

    /**
     * Execute the query and return the filtered Set.
     * The query is executed with the parameters set by the Map values.  Each
     * Map entry consists of a key which is the name of the parameter in the
     * declareParameters method, and a value which is the value used in
     * the execute method.  The keys in the Map and the declared parameters
     * must exactly match or a JDOUserException is thrown.
     * @return the filtered Set.
     * @see #executeWithArray(Object[] parameters)
     * @param amap the Map containing all of the parameters.
     */
    public Object executeWithMap(Map amap) {
        beforeExecute(amap.size());
        try {
            Object res = qc.execute(amap, pm);
            result.add(res);
            return res;
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "execute failed", e);
            throw new JDOException("Impossible to execute a query ", e);
        }
    }

    /**
     * Execute the query and return the filtered Set.
     * <P>The execution of the query obtains the values of the parameters and
     * matches them against the declared parameters in order.  The names
     * of the declared parameters are ignored.  The type of
     * the declared parameters must match the type of the passed parameters,
     * except that the passed parameters might need to be unwrapped to get
     * their primitive values.
     *
     * <P>The filter, import, declared parameters, declared variables, and
     * ordering statements are verified for consistency.
     *
     * <P>Each element in the candidate Set is examined to see that it
     * is assignment compatible to the Class of the query.  It is then evaluated
     * by the boolean expression of the filter.  The element passes the filter
     * if there exist unique values for all variables for which the filter
     * expression evaluates to true.
     * @return the filtered Set.
     * @param anarray the Object array with all of the parameters.
     */
    public Object executeWithArray(Object[] anarray) {
        beforeExecute(anarray.length);
        try {
            return qc.execute(anarray, pm);
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR, "execute failed", e);
            throw new JDOException("Impossible to execute a query ", e);
        }
    }

    /**
     * Get the PersistenceManager associated with this Query.
     * <P>If this Query was restored from a serialized form, it has no
     * PersistenceManager, and this method returns null.
     * @return the PersistenceManager associated with this Query.
     */
    public PersistenceManager getPersistenceManager() {
        return pm;
    }

    /**
     * Close a query result and release any resources
     * associated with it.  The parameter is the return from execute(...) and
     * might have iterators open on it. Iterators associated with the query
     * result are invalidated: they return false to hasNext() and throw
     * NoSuchElementException to next().
     *
     * This method makes nothing, all is done by the SpeedoQueryCompiler
     * object. This object is never closed, and could be reused in the
     * future.
     *
     * @param queryResult the result of execute(...) on this Query instance.
     */
    public void close(Object queryResult) {
        ((QueryResultCollection) queryResult).close();
        int i=0;
        while(i < result.size() && result.get(i) != queryResult) {
            i++;
        }
        if (i < result.size()) {
            result.remove(i);
            if (result.size() == 0) {
                try {
                    pool.releaseResource(this);
                } catch (Exception e) {
                    logger.log(BasicLevel.ERROR,
                        "Impossible to release a SpeedoQuery into the pool", e);
                }
            }
        }
    }

    /**
     * Close all query results associated with this Query
     * instance, and release all resources associated with them.  The query
     * results might have iterators open on them.  Iterators associated with the
     * query results are invalidated: they return false to hasNext() and throw
     * NoSuchElementException to next().
     *
     * This method makes nothing, all is done by the SpeedoQueryCompiler
     * object. This object is never closed, and could be reused in the
     * future.
     *
     */
    public void closeAll() {
        for(int i=0; i<result.size(); i++) {
            ((QueryResultCollection) result.get(i)).close();
        }
        result.clear();
        try {
            pool.releaseResource(this);
        } catch (Exception e) {
            logger.log(BasicLevel.ERROR,
                "Impossible to release a SpeedoQuery into the pool", e);
        }
    }

    // IMPLEMENTATION OF THE QueryDefinition INTERFACE //
    //-------------------------------------------------//

    public Class getCandidateClass() {
        return candidateClass;
    }

    public boolean isCollection() {
        return (candidateInstances != null);
    }

    public Collection getCollection() {
        return candidateInstances;
    }

    public String getParameters() {
        return parameters;
    }

    public String getVariables() {
        return variables;
    }

    public String getFilter() {
        return filter;
    }

    public Extent getExtent() {
        return extentClass;
    }

    public List getImportStatements() {
        return importStatements;
    }

    public boolean isIgnoreCache() {
        return ignoreCache;
    }

    public List getOrder() {
        return order;
    }

    // PRIVATE METHODS //
    //-----------------//

    private void beforeExecute(int nbparameters) {
        if (parameters != null && nbparameters == 0)
            throw new JDOUserException(
                    "Parameters in not null, and you try to execute without parameters");
        if (pm.isClosed())
            throw new JDOFatalUserException(
                    "Impossible to use the query: the persistent manager is closed.");

        if (hasChanged) {
            compile();
        }
    }

    private void assertPMIsOpen() {
        if (pm.isClosed())
            throw new JDOUserException("The persistence manager is closed.");
    }

    private void assertParameters() {
        if (parameters == null)
            throw new JDOUserException("Parameters string can not be null");
    }

    private void assertVariables() {
        if (variables == null)
            throw new JDOUserException("variables string can not be null");
    }

    private void assertCandidateClass() {
        if (candidateClass == null)
            throw new JDOUserException("candidate class can not be null");
    }

    private void assertExtentClass() {
        if (extentClass == null)
            throw new JDOUserException("Extent class can not be null");
    }

    private void assertCandidateInstances() {
        if (candidateInstances == null)
            throw new JDOUserException("candidate Collection can not be null");
    }

}
