/**
 * Speedo: an implementation of JDO compliant personality on top of JORM generic
 * I/O sub-system.
 * Copyright (C) 2001-2003 France Telecom R&D
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Release: 0.1
 *
 * Contact: pascal.dechamboux@rd.francetelecom.com
 *
 * Authors: X. Spengler
 *
 */

package org.objectweb.speedo.runtime.query.parser;

import org.objectweb.jorm.metainfo.api.Class;
import org.objectweb.jorm.metainfo.api.Manager;
import org.objectweb.medor.api.Field;
import org.objectweb.medor.query.api.PropagatedField;
import org.objectweb.medor.api.MedorException;
import org.objectweb.medor.filter.lib.BasicFieldOperand;
import org.objectweb.medor.filter.lib.NavigatorOperator;
import org.objectweb.medor.query.api.QueryTree;
import org.objectweb.medor.query.jorm.lib.ClassExtent;
import org.objectweb.medor.query.lib.Cartesian;
import org.objectweb.medor.query.jorm.lib.NavigatorNodeFactory;
import org.objectweb.medor.query.jorm.lib.PNameField;
import org.objectweb.util.monolog.api.BasicLevel;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Stack;
import java.util.Map;

/**
 * Implementation of a visitor that creates a map of pairs [id,JormExtent] for all defined
 * identifiers of the query.
 */
public class SpeedoQLVariableVisitor extends SpeedoQLAbstractVisitor {

    /**
     * manager is the Jorm meta information manager used  to search the
     * appropriate class
     */
    private Manager manager;

    /**
     * ids is a temporary structure to store variables, parameters, name
     * definitions, and build corresponding QueryTree(s)
     * key: the name to resolve
     * value: an IdValue object which contains the name, the paths, and the
     * corresponding QueryTree (see the definition of IdValue object in
     * SpeedoQLAbstractVisitor object)
     */
    private Map ids = new HashMap();

    /**
     * qt is the built QueryTree from the current object.
     */
    private QueryTree qt = null;

    /**
     * speedoql is the result of the filter parsing
     */
    private SimpleNode speedoql = null;

    public ClassExtent classExtent = null;


    /**
     * constructor
     * @param speedoql root of the lexical tree
     * @param manager JORM meta information manager
     */
    public SpeedoQLVariableVisitor(SimpleNode speedoql, Manager manager) {
        this.speedoql = speedoql;
        this.manager = manager;
    }

    /**
     * The visit of the tree starts here.
     * Please setup current class, params and vars hashtable before calling
     * this method.
     */
    public void startVisiting() throws Exception {

        // first step, we need to add "this" in key and the current class as
        // the first IdValue object
        if (curClass != null) {
            logger.log(BasicLevel.DEBUG, "create a new IdValue object " +
                    "with the current class (" + curClass + ")");

            IdValue iv = new IdValue();
            iv.name = new String[1];
            iv.name[0] = curClass;
            iv.nameType = SpeedoQLAbstractVisitor.JORM_NAME;
            ids.put("this", iv);
        }

        // first step, visit the tree
        visit(speedoql);

        for (Iterator i = ids.keySet().iterator(); i.hasNext();) {
            String key = (String) i.next();
            logger.log(BasicLevel.DEBUG, "key=<" + key + "> [" + ids.get(key) + "]");
        }

        // replace map values with JormExtent or Medor Field
        for (Iterator i = ids.keySet().iterator(); i.hasNext();) {
            String iter = (String) i.next();
            getQueryTree(iter);
        }

        for (Iterator i = ids.keySet().iterator(); i.hasNext();) {
            String key = (String) i.next();
            logger.log(BasicLevel.DEBUG, "key=<" + key + "> [" + ids.get(key) + "]");
        }

        HashSet s = new HashSet();
        Iterator i = null;
        for (i = ids.values().iterator(); i.hasNext();) {
            s.add(((IdValue) i.next()).qt);
        }

        i = s.iterator();
        qt = (QueryTree) i.next();
        while (i.hasNext()) {
            qt = new Cartesian(qt, (QueryTree) i.next(), null, null);
        }
    }

    private QueryTree getQueryTree(String id) throws ParseException, MedorException {
        logger.log(BasicLevel.DEBUG, "\nQueryTree id=(" + id + ")");

        IdValue iv = null;

        if (id.compareTo(curClass) == 0) {
            iv = (IdValue) ids.get("this");
        } else
            iv = (IdValue) ids.get(id);

        if ((iv == null) && (id.compareTo(curClass) != 0))
            throw new ParseException("Abstract schema name \""
                    + id + "\" has not been declared");

        if (iv.name == null)
            throw new ParseException("Identifer '" + id + "' has not been declared");

        if (iv.qt != null)
            return iv.qt;

        if (iv.name.length == 1) {
            // Abstract schema name case
            Class theClass = manager.getClass(iv.name[0]);
            if (theClass == null) {
                throw new ParseException("Abstract schema name \"" + iv.name[0]
                        + "\" has not been declared in the jorm meta information");
            }
            classExtent = new ClassExtent(
                theClass, curClass, Field.PNAMENAME, false);
            iv.qt = classExtent;

        } else {
            iv.qt = getQueryTree(iv.name[0]);
            iv.qt = getNavNode(iv.qt, iv.name, id, iv.nameType, false);
        }

        for (int i = 0; i < iv.getDeclaredPathLength(); i++) {
            String[] p = iv.getSplitedPath(i);
            //logger.log(BasicLevel.DEBUG, "PATH["+i+"]="+iv.getMergedPath(i)+";");
            iv.qt = getNavNode(iv.qt, p, mergePath(p, 0, p.length - 1), iv.nameType, true);
        }

        if (iv.name.length > 1)
            setQueryTree(iv.name[0], iv.qt);

        return iv.qt;
    }

    private QueryTree getNavNode(QueryTree qt,
                                 String[] pathName,
                                 String alias,
                                 int name_t,
                                 boolean isPath) throws ParseException {
        logger.log(BasicLevel.DEBUG, "enter->getNavNode=<" + alias + ">");
        try {
            NavigatorOperator nav = null;
            Field f = null;

            switch(name_t) {
            case SpeedoQLAbstractVisitor.NAVIGATION:
            case SpeedoQLAbstractVisitor.JORM_NAME:
                // navigation case and JORM name use
                logger.log(BasicLevel.DEBUG, "<" + alias + ">NAVIGATION or JORM_NAME");
                logger.log(BasicLevel.DEBUG, "Field=[" + curClass + "." + Field.PNAMENAME + "]");
                f = qt.getTupleStructure().getField(curClass + "." + Field.PNAMENAME);
                // create a first NavigatorOperator with a Field
                String build = curClass;
                logger.log(BasicLevel.DEBUG, "f.name: " + f.getName() + " / nav(" + pathName[1] + ") / f:" + f);
                logger.log(BasicLevel.DEBUG, "qt: " + qt);
                PNameField pnf = null;
                if (f instanceof PNameField) {
                    pnf = (PNameField) f;
                } else if (f instanceof PropagatedField) {
                    pnf = (PNameField) ((PropagatedField) f).getOriginFields()[0];
                }
                Class c = pnf.getMetaObjectClass();
                if (c.getTypedElement(pathName[1]) == null) {
                    throw new ParseException("The field " + pathName[1]
                            + " is not defined in the class " + c.getFQName());
                }
                nav = new NavigatorOperator(new BasicFieldOperand(f), pathName[1]);
                build += "." + pathName[1];
                logger.log(BasicLevel.DEBUG, "LENGTH=[" + pathName.length + "]");

                // go to recursively add all elements after '.'
                for (int i = 2; i <= pathName.length - 1; i++) {
                    logger.log(BasicLevel.DEBUG, "Navigator=[" + i + "][+" + pathName[i] + "]");
                    nav = new NavigatorOperator(nav, pathName[i]);
                    if (i < pathName.length - 1)
                        build += "." + pathName[i];
                }

                //logger.log(BasicLevel.DEBUG, "ALIAS=[" + alias+ "]+["+alias2fullname(alias)+"]");
                //return NavigatorNodeFactory.getNavigatorNode(nav, alias2fullname(alias), null);
                logger.log(BasicLevel.DEBUG, "ALIAS=[" + alias + "]+[" + build + "]");

                return NavigatorNodeFactory.getNavigatorNode(nav, build, null);

            case SpeedoQLAbstractVisitor.CONTAINS_IN:
                // Collections.contains(aVariable); --> getIn
                logger.log(BasicLevel.DEBUG, "CONTAINS_IN: ");

                if (!isPath) {

                    logger.log(BasicLevel.DEBUG, "make a nav with:(" + pathName.length + ")" + pathName[0] + "-" + pathName[1] + " alias=(" + alias + ")");

                    //Field f = qt.getTupleStructure().getField(getEndString(pathName[0]) + "." + Field.PNAMENAME);
                    f = qt.getTupleStructure().getField(pathName[0] + "." + Field.PNAMENAME);
                    nav = new NavigatorOperator(new BasicFieldOperand(f), pathName[1]);

                    logger.log(BasicLevel.DEBUG, "getIN=<" + alias + ">");
                    return NavigatorNodeFactory.getIn(nav, null, alias);

                } else {
                    logger.log(BasicLevel.DEBUG, "make a nav with(else):(" + pathName.length + ")" + pathName[0] + "-" + pathName[1] + " alias=(" + alias + ")");

                    f = qt.getTupleStructure().getField(alias);
                    nav = new NavigatorOperator(new BasicFieldOperand(f), pathName[1]);
                    return NavigatorNodeFactory.getNavigatorNode(nav, alias, null);
                }
            case SpeedoQLAbstractVisitor.CONTAINS_MEMBEROF:
                // Collections.contains(aParameter); --> getMemberOf
                logger.log(BasicLevel.DEBUG, "CONTAINS_MEMBEROF: NYI");
                // with a parameter (nav, FieldOperand, omd=null???)
                //return NavigatorNodeFactory.getMemberOf(nav, null, null);
                return null;
            default:
                return null;
            }
        } catch (MedorException e) {
            // just print the stack for the moment
            e.printStackTrace();
            throw new VisitorException(e);
        }
    }

    private void setQueryTree(String id, QueryTree aqt) {
        IdValue iv = null;
        if (id.compareTo(curClass) == 0)
            iv = (IdValue) ids.get("this");
        else
            iv = (IdValue) ids.get(id);

        iv.qt = aqt;
        if (iv.name.length > 1)
            setQueryTree(iv.name[0], iv.qt);
    }

    /**
     * get the Map that was built from visiting the lexical query tree
     */
    public QueryTree getQueryTree() {
        return qt;
    }


    /**
     * ********************* VISITOR METHODS ***********************************
     */

    public Object visit(ASTSpeedoPrimary node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTSpeedoQL node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTPrimary node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTRelationalExpression node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTAdditiveExpression node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTUnaryExpression node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTCastExpression node, Object data) {
        return null;
    }

    public Object visit(ASTArgumentList node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTLiteral node, Object data) {
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTType node, Object data) {
        //voir cast , donc on s'en fout
        visit((SimpleNode) node, data);
        return null;
    }

    public Object visit(ASTQualifiedName node, Object data) {

        String name = (String) node.value;
        logger.log(BasicLevel.DEBUG, "variable visitor: qualifiedname=<" + name + ">");

        String[] splitted = splitPath(name);
        String result = null;
        IdValue iv = null;
        String prefix = null;
        if ((params != null) && (params.get(splitted[0]) != null)) {
            // tests with the params table
            logger.log(BasicLevel.DEBUG, "in PARAM");

/*
if (((Stack) data).size() > 0) {
String dataStuff = (String) ((Stack) data).pop();
String[] dataStuffSplitted = splitPath(dataStuff);
iv = (IdValue) ids.get(name);
if (iv == null) {
if ((params != null) && (params.get(dataStuffSplitted[0]) != null)) {
// dont know if it is possible
} else if ((vars != null) && ((vars.get(dataStuffSplitted[0])) != null)) {
// dont know if it is possible
} else {
// class field
String classField = curClass+"."+dataStuff.substring(0,dataStuff.lastIndexOf("."));
logger.log(BasicLevel.DEBUG, "classField=<"+classField+">");

iv = new IdValue(classField);
ids.put(name, iv);
}

} else { //iv != null
iv.addPath(name);
}
iv.nameType = SpeedoQLAbstractVisitor.CONTAINS_MEMBEROF;
}
*/
            /*
            else { // stack size = 0
            iv = (IdValue) ids.get(name);
            if (iv == null) {
            iv = new IdValue();
            iv.addPath(name);
            ids.put(splitted[0], iv);
            } else {
            iv.addPath(name);
            }
            }
            iv.nameType = SpeedoQLAbstractVisitor.CONTAINS_MEMBEROF;
*/

        } else if ((vars != null) && ((vars.get(splitted[0])) != null)) {
            logger.log(BasicLevel.DEBUG, "in VAR");
            if (((Stack) data).size() > 0) {
                logger.log(BasicLevel.DEBUG, "in VAR stack > 0");
                String dataStuff = (String) ((Stack) data).pop();
                logger.log(BasicLevel.DEBUG, "in VAR stack > 0 dataStuff=<" + dataStuff + ">");
                String[] dataStuffSplitted = splitEndPath(dataStuff, curClass);
                iv = (IdValue) ids.get(splitted[0]);
                if (iv == null) {
                    logger.log(BasicLevel.DEBUG, "in VAR iv null");
                    logger.log(BasicLevel.DEBUG, "create a new IdValue with name=<" + dataStuff + ">, key=<" + splitted[0] + ">");

                    iv = new IdValue(dataStuff);
                    ids.put(splitted[0], iv);

                } else { //iv != null
                    logger.log(BasicLevel.DEBUG, "in VAR iv != null iv.name='" + dataStuffSplitted + "'");
                    for (int k = 0; k < dataStuffSplitted.length; k++) {
                        logger.log(BasicLevel.DEBUG, "[" + k + "]='" + dataStuffSplitted[k] + "'");
                    }
                    iv.name = dataStuffSplitted;

//					IdValue iv2 = (IdValue) (ids.get(dataStuffSplitted[0]));
/*
					if (iv2 != null) { // all CASES
						logger.log(BasicLevel.DEBUG, "====="+dataStuff+"========");
						iv2.addPath(dataStuff);
					} else
						logger.log(BasicLevel.DEBUG, "NOT POSSIBLE !!!!!!!!!!!!!!");
*/

                }
            } else { // stack size = 0
                logger.log(BasicLevel.DEBUG, "in VAR stack == 0");
                iv = (IdValue) ids.get(splitted[0]);
                if (iv == null) {
                    logger.log(BasicLevel.DEBUG, "in VAR stack == 0, iv null");
                    iv = new IdValue();
                    iv.addPath(name);
                    ids.put(splitted[0], iv);
                } else {
                    logger.log(BasicLevel.DEBUG, "in VAR stack == 0, iv != null======(" + name + ")");
                    iv.addPath(name);
                }
            }
            iv.nameType = SpeedoQLAbstractVisitor.CONTAINS_IN;

        } else { // current class
            logger.log(BasicLevel.DEBUG, "in CURRENT CLASS:");

            String trunked;
            if ("contains".compareTo(splitted[splitted.length - 1]) == 0)
                trunked = curClass + "." + name.substring(0, name.lastIndexOf("."));
            else
                trunked = curClass + "." + name;
            /*
            if ((name.lastIndexOf(".") != -1) && ("contains".compareTo(splitted[splitted.length -1]) == 0))
            trunked = curClass+"."+name.substring(0, name.lastIndexOf("."));
            else
            trunked = curClass+"."+name;
*/

            iv = (IdValue) ids.get("this");
            //iv.addPath(trunked);

            if ("contains".compareTo(splitted[splitted.length - 1]) == 0) {
                logger.log(BasicLevel.DEBUG, "contains case: addPath=<" + trunked + ">");
                //stock in stack a mean to find the contains stuff...

                ((Stack) data).push(trunked);
            } else {
                // we know that we are on a NAVIGATION case
                //logger.log(BasicLevel.DEBUG, "Navigation case, addpath=<" + trunked + ">");
                logger.log(BasicLevel.DEBUG, "Navigation case, addpath=<this." + name + ">");
                iv.nameType = SpeedoQLAbstractVisitor.NAVIGATION;
                //iv.addPath(trunked);
                if (name.startsWith("this"))
                    iv.addPath(name);
                else
                    iv.addPath("this." + name);
            }

        }

        logger.log(BasicLevel.DEBUG, "variable visitor: qualifiedname=<" + name + ">-end");
        return null;
    }

}
